# Copyright (c) 2015 Red Hat Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import os

import eventlet
import mock
import netaddr
from oslo_config import fixture as fixture_config
from oslo_utils import uuidutils

from neutron.agent.l3 import keepalived_state_change
from neutron.agent.linux import ip_lib
from neutron.common import utils
from neutron.conf.agent.l3 import keepalived as kd
from neutron.tests.common import machine_fixtures as mf
from neutron.tests.common import net_helpers
from neutron.tests.functional import base


def has_expected_arp_entry(device_name, namespace, ip, mac):
    ip_version = utils.get_ip_version(ip)
    entry = ip_lib.dump_neigh_entries(ip_version, device_name, namespace,
                                      dst=ip, lladdr=mac)
    return entry != []


class TestKeepalivedStateChange(base.BaseSudoTestCase):
    def setUp(self):
        super(TestKeepalivedStateChange, self).setUp()
        self.conf_fixture = self.useFixture(fixture_config.Config())
        kd.register_l3_agent_keepalived_opts(self.conf_fixture)
        self.router_id = uuidutils.generate_uuid()
        self.conf_dir = self.get_default_temp_dir().path
        self.cidr = '169.254.128.1/24'
        self.interface_name = utils.get_rand_name()
        self.monitor = keepalived_state_change.MonitorDaemon(
            self.get_temp_file_path('monitor.pid'),
            self.router_id,
            1,
            2,
            utils.get_rand_name(),
            self.conf_dir,
            self.interface_name,
            self.cidr)
        mock.patch.object(self.monitor, 'notify_agent').start()
        self.line = '1: %s    inet %s' % (self.interface_name, self.cidr)

    def test_parse_and_handle_event_wrong_device_completes_without_error(self):
        self.monitor.parse_and_handle_event(
            '1: wrong_device    inet wrong_cidr')

    def _get_state(self):
        with open(os.path.join(self.monitor.conf_dir, 'state')) as state_file:
            return state_file.read()

    def test_parse_and_handle_event_writes_to_file(self):
        self.monitor.parse_and_handle_event('Deleted %s' % self.line)
        self.assertEqual('backup', self._get_state())

        self.monitor.parse_and_handle_event(self.line)
        self.assertEqual('master', self._get_state())

    def test_parse_and_handle_event_fails_writing_state(self):
        with mock.patch.object(
                self.monitor, 'write_state_change', side_effect=OSError):
            self.monitor.parse_and_handle_event(self.line)

    def test_parse_and_handle_event_fails_notifying_agent(self):
        with mock.patch.object(
                self.monitor, 'notify_agent', side_effect=Exception):
            self.monitor.parse_and_handle_event(self.line)

    def test_handle_initial_state_backup(self):
        ip = ip_lib.IPWrapper(namespace=self.monitor.namespace)
        ip.netns.add(self.monitor.namespace)
        self.addCleanup(ip.netns.delete, self.monitor.namespace)
        ip.add_dummy(self.interface_name)

        with mock.patch.object(
                self.monitor, 'write_state_change') as write_state_change,\
                mock.patch.object(
                    self.monitor, 'notify_agent') as notify_agent:

            self.monitor.handle_initial_state()
            write_state_change.assert_not_called()
            notify_agent.assert_not_called()

    def test_handle_initial_state_master(self):
        ip = ip_lib.IPWrapper(namespace=self.monitor.namespace)
        ip.netns.add(self.monitor.namespace)
        self.addCleanup(ip.netns.delete, self.monitor.namespace)
        ha_interface = ip.add_dummy(self.interface_name)

        ha_interface.addr.add(self.cidr)

        self.monitor.handle_initial_state()
        self.assertEqual('master', self._get_state())


class TestMonitorDaemon(base.BaseSudoTestCase):
    def setUp(self):
        super(TestMonitorDaemon, self).setUp()
        bridge = self.useFixture(net_helpers.OVSBridgeFixture()).bridge
        self.machines = self.useFixture(mf.PeerMachines(bridge))
        self.router, self.peer = self.machines.machines[:2]

        conf_dir = self.get_default_temp_dir().path
        monitor = keepalived_state_change.MonitorDaemon(
            self.get_temp_file_path('monitor.pid'),
            uuidutils.generate_uuid(),
            1,
            2,
            self.router.namespace,
            conf_dir,
            'foo-iface',
            self.machines.ip_cidr
        )
        eventlet.spawn_n(monitor.run, run_as_root=True)
        monitor_started = functools.partial(
            lambda mon: mon.monitor is not None, monitor)
        utils.wait_until_true(monitor_started)
        self.addCleanup(monitor.monitor.stop)

    def test_new_fip_sends_garp(self):
        next_ip_cidr = net_helpers.increment_ip_cidr(self.machines.ip_cidr, 2)
        expected_ip = str(netaddr.IPNetwork(next_ip_cidr).ip)
        # Create incomplete ARP entry
        self.peer.assert_no_ping(expected_ip)
        has_entry = has_expected_arp_entry(
            self.peer.port.name,
            self.peer.namespace,
            expected_ip,
            self.router.port.link.address)
        self.assertFalse(has_entry)
        self.router.port.addr.add(next_ip_cidr)
        has_arp_entry_predicate = functools.partial(
            has_expected_arp_entry,
            self.peer.port.name,
            self.peer.namespace,
            expected_ip,
            self.router.port.link.address,
        )
        exc = RuntimeError(
            "No ARP entry in %s namespace containing IP address %s and MAC "
            "address %s" % (
                self.peer.namespace,
                expected_ip,
                self.router.port.link.address))
        utils.wait_until_true(
            has_arp_entry_predicate,
            exception=exc)



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""device_owner_ha_replicate_int

Revision ID: 7bbb25278f53
Revises: 4ffceebfcdc
Create Date: 2016-03-22 10:00:43.245503

"""

# revision identifiers, used by Alembic.
revision = '7bbb25278f53'
down_revision = '4ffceebfcdc'

from alembic import op
from neutron_lib import constants
import sqlalchemy as sa


ROUTER_ATTR_TABLE = 'router_extra_attributes'
ROUTER_PORTS_TABLE = 'routerports'
PORTS_TABLE = 'ports'


def upgrade():
    update_device_owner_ha_replicated_interface()


def update_device_owner_ha_replicated_interface():
    router_attr_table = sa.Table(ROUTER_ATTR_TABLE,
                                 sa.MetaData(),
                                 sa.Column('router_id', sa.String(36)),
                                 sa.Column('ha', sa.Boolean),)

    routerports = sa.Table(ROUTER_PORTS_TABLE,
                           sa.MetaData(),
                           sa.Column('router_id', sa.String(36)),
                           sa.Column('port_type', sa.String(255)))

    ports = sa.Table(PORTS_TABLE,
                     sa.MetaData(),
                     sa.Column('device_owner', sa.String(255)),
                     sa.Column('device_id', sa.String(255)))

    session = sa.orm.Session(bind=op.get_bind())
    with session.begin(subtransactions=True):
        for router_attr in session.query(
                router_attr_table).filter(router_attr_table.c.ha):
            session.execute(routerports.update().values(
                port_type=constants.DEVICE_OWNER_HA_REPLICATED_INT).where(
                routerports.c.router_id == router_attr.router_id).where(
                routerports.c.port_type == constants.DEVICE_OWNER_ROUTER_INTF))
            session.execute(ports.update().values(
                device_owner=constants.DEVICE_OWNER_HA_REPLICATED_INT).where(
                ports.c.device_id == router_attr.router_id).where(
                ports.c.device_owner == constants.DEVICE_OWNER_ROUTER_INTF))
    session.commit()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
import sqlalchemy as sa
from sqlalchemy import orm

from neutron.db import models_v2


class PortSecurityBinding(model_base.BASEV2):
    port_id = sa.Column(sa.String(36),
                        sa.ForeignKey('ports.id', ondelete="CASCADE"),
                        primary_key=True)
    port_security_enabled = sa.Column(sa.Boolean(), nullable=False)

    # Add a relationship to the Port model in order to be to able to
    # instruct SQLAlchemy to eagerly load port security binding
    port = orm.relationship(
        models_v2.Port, load_on_pending=True,
        backref=orm.backref("port_security", uselist=False,
                            cascade='delete', lazy='joined'))
    revises_on_change = ('port',)


class NetworkSecurityBinding(model_base.BASEV2):
    network_id = sa.Column(sa.String(36),
                           sa.ForeignKey('networks.id', ondelete="CASCADE"),
                           primary_key=True)
    port_security_enabled = sa.Column(sa.Boolean(), nullable=False)

    # Add a relationship to the Port model in order to be able to instruct
    # SQLAlchemy to eagerly load default port security setting for ports
    # on this network
    network = orm.relationship(
        models_v2.Network, load_on_pending=True,
        backref=orm.backref("port_security", uselist=False,
                            cascade='delete', lazy='joined'))
    revises_on_change = ('network',)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import oslo_messaging
from oslo_utils import uuidutils
import testtools

from neutron.api.rpc.callbacks import events
from neutron.api.rpc.handlers import resources_rpc
from neutron.objects import trunk
from neutron.services.trunk import constants as t_const
from neutron.services.trunk.drivers.linuxbridge.agent import driver
from neutron.services.trunk.drivers.linuxbridge.agent import trunk_plumber
from neutron.tests import base


class LinuxBridgeTrunkDriverTestCase(base.BaseTestCase):
    def setUp(self):
        super(LinuxBridgeTrunkDriverTestCase, self).setUp()
        self.plumber = mock.create_autospec(trunk_plumber.Plumber())
        self.stub = mock.create_autospec(driver.trunk_rpc.TrunkStub())
        self.tapi = mock.create_autospec(driver._TrunkAPI(self.stub))
        self.lbd = driver.LinuxBridgeTrunkDriver(self.plumber, self.tapi)
        self.trunk = trunk.Trunk(id=uuidutils.generate_uuid(),
                                 port_id=uuidutils.generate_uuid(),
                                 project_id=uuidutils.generate_uuid())
        self.subports = [trunk.SubPort(id=uuidutils.generate_uuid(),
                                       port_id=uuidutils.generate_uuid(),
                                       segmentation_type='vlan',
                                       trunk_id=self.trunk.id,
                                       segmentation_id=i)
                         for i in range(20)]
        self.trunk.sub_ports = self.subports

    def test_handle_trunks_created(self):
        self._test_handle_trunks_wire_event(events.CREATED)

    def test_handle_trunks_updated(self):
        self._test_handle_trunks_wire_event(events.UPDATED)

    def _test_handle_trunks_wire_event(self, event):
        self.plumber.trunk_on_host.return_value = True
        self.lbd.handle_trunks(mock.Mock(), 'TRUNKS',
                               [self.trunk], event)
        self.tapi.put_trunk.assert_called_once_with(
            self.trunk.port_id, self.trunk)
        self.tapi.bind_subports_to_host.assert_called_once_with(
            mock.ANY, self.trunk)
        self.assertFalse(self.plumber.delete_trunk_subports.called)

    def test_handle_trunks_deleted(self):
        self.lbd.handle_trunks(mock.Mock(), 'TRUNKS',
                               [self.trunk], events.DELETED)
        self.tapi.put_trunk.assert_called_once_with(
            self.trunk.port_id, None)
        self.plumber.delete_trunk_subports.assert_called_once_with(self.trunk)

    def test_handle_subports_deleted(self):
        self.tapi.get_trunk_by_id.return_value = self.trunk
        self.lbd.handle_subports(mock.Mock(), 'TRUNKS',
                                 self.trunk.sub_ports, events.DELETED)
        self.assertEqual(20, len(self.tapi.delete_trunk_subport.mock_calls))
        # should have tried to wire trunk at the end with state
        self.plumber.trunk_on_host.assert_called_once_with(self.trunk)

    def test_handle_subports_created(self):
        self.tapi.get_trunk_by_id.return_value = self.trunk
        self.lbd.handle_subports(mock.Mock(), 'TRUNKS',
                                 self.trunk.sub_ports, events.CREATED)
        self.assertEqual(20, len(self.tapi.put_trunk_subport.mock_calls))
        # should have tried to wire trunk at the end with state
        self.plumber.trunk_on_host.assert_called_once_with(self.trunk)

    def test_agent_port_change_is_trunk(self):
        self.tapi.get_trunk.return_value = self.trunk
        self.lbd.agent_port_change('resource', 'event', 'trigger', 'context',
                                   {'port_id': self.trunk.port_id})
        # should have tried to wire trunk
        self.plumber.trunk_on_host.assert_called_once_with(self.trunk)

    def test_agent_port_change_not_trunk(self):
        self.tapi.get_trunk.return_value = None
        self.tapi.get_trunk_for_subport.return_value = None
        other_port_id = uuidutils.generate_uuid()
        self.lbd.agent_port_change('resource', 'event', 'trigger', 'context',
                                   {'port_id': other_port_id})
        self.plumber.delete_subports_by_port_id.assert_called_once_with(
            other_port_id)

    def test_agent_port_change_is_subport(self):
        self.tapi.get_trunk.return_value = None
        self.tapi.get_trunk_for_subport.return_value = self.trunk
        self.lbd.agent_port_change('resource', 'event', 'trigger', 'context',
                                   {'port_id': self.trunk.sub_ports[0].port_id,
                                    'mac_address': 'mac_addr'})
        self.plumber.delete_subports_by_port_id.assert_called_once_with(
            self.trunk.sub_ports[0].port_id)

    def test_wire_trunk_happy_path(self):
        self.lbd.wire_trunk('ctx', self.trunk)
        self.tapi.bind_subports_to_host.assert_called_once_with(
            'ctx', self.trunk)
        self.plumber.ensure_trunk_subports.assert_called_once_with(self.trunk)
        self.tapi.set_trunk_status.assert_called_once_with(
            'ctx', self.trunk, t_const.ACTIVE_STATUS)

    def test_wire_trunk_not_on_host(self):
        # trunk device not on host
        self.plumber.trunk_on_host.return_value = False
        self.lbd.wire_trunk('ctx', self.trunk)
        # don't bind and don't set status
        self.assertFalse(self.tapi.bind_subports_to_host.called)
        self.assertFalse(self.tapi.set_trunk_status.called)

    def test_wire_trunk_concurrent_removal(self):
        self.plumber.trunk_on_host.side_effect = [True, False]
        self.plumber.ensure_trunk_subports.side_effect = ValueError()
        self.lbd.wire_trunk('ctx', self.trunk)
        # we don't change status if port was just removed
        self.assertFalse(self.tapi.set_trunk_status.called)

    def test_wire_trunk_other_exception(self):
        self.plumber.ensure_trunk_subports.side_effect = ValueError()
        self.lbd.wire_trunk('ctx', self.trunk)
        # degraded due to dataplane failure
        self.tapi.set_trunk_status.assert_called_once_with(
            'ctx', self.trunk, t_const.DEGRADED_STATUS)


class TrunkAPITestCase(base.BaseTestCase):
    def setUp(self):
        super(TrunkAPITestCase, self).setUp()
        self.stub = mock.create_autospec(driver.trunk_rpc.TrunkStub())
        self.tapi = driver._TrunkAPI(self.stub)
        self.trunk = trunk.Trunk(id=uuidutils.generate_uuid(),
                                 port_id=uuidutils.generate_uuid(),
                                 project_id=uuidutils.generate_uuid())
        self.subports = [trunk.SubPort(id=uuidutils.generate_uuid(),
                                       port_id=uuidutils.generate_uuid(),
                                       segmentation_type='vlan',
                                       trunk_id=self.trunk.id,
                                       segmentation_id=i)
                         for i in range(20)]
        self.trunk.sub_ports = self.subports
        self.stub.get_trunk_details.return_value = self.trunk

    def test_fetch_trunk(self):
        self.assertEqual(self.trunk, self.tapi._fetch_trunk('ctx', 'port'))
        self.stub.get_trunk_details.assert_called_once_with('ctx', 'port')

    def test_fetch_trunk_missing(self):
        self.stub.get_trunk_details.side_effect = (
            resources_rpc.ResourceNotFound(resource_id='1', resource_type='1'))
        self.assertIsNone(self.tapi._fetch_trunk('ctx', 'port'))

    def test_fetch_trunk_plugin_disabled(self):
        self.stub.get_trunk_details.side_effect = (
            oslo_messaging.RemoteError('CallbackNotFound'))
        self.assertIsNone(self.tapi._fetch_trunk('ctx', 'port'))

    def test_fetch_trunk_plugin_other_error(self):
        self.stub.get_trunk_details.side_effect = (
            oslo_messaging.RemoteError('vacuum full'))
        with testtools.ExpectedException(oslo_messaging.RemoteError):
            self.tapi._fetch_trunk('ctx', 'port')

    def test_set_trunk_status(self):
        self.tapi.set_trunk_status('ctx', self.trunk, 'STATUS')
        self.stub.update_trunk_status.assert_called_once_with(
            'ctx', self.trunk.id, 'STATUS')

    def test_bind_subports_to_host(self):
        self.tapi.bind_subports_to_host('ctx', self.trunk)
        self.stub.update_subport_bindings.assert_called_once_with(
            'ctx', self.trunk.sub_ports)

    def test_put_trunk_subport_non_existent_trunk(self):
        # trunks not registered are ignored
        self.tapi.put_trunk_subport(
            'non_trunk_id', self.trunk.sub_ports[0])

    def test_get_trunk_by_id(self):
        self.tapi.put_trunk(self.trunk.port_id, self.trunk)
        self.assertEqual(self.trunk,
                         self.tapi.get_trunk_by_id('ctx', self.trunk.id))
        self.assertIsNone(self.tapi.get_trunk_by_id('ctx', 'other_id'))

    def test_put_trunk_subport(self):
        self.tapi.put_trunk(self.trunk.port_id, self.trunk)
        new = trunk.SubPort(id=uuidutils.generate_uuid(),
                            port_id=uuidutils.generate_uuid(),
                            segmentation_type='vlan',
                            trunk_id=self.trunk.id,
                            segmentation_id=1010)
        self.tapi.put_trunk_subport(self.trunk.id, new)
        subs = self.tapi.get_trunk('ctx', self.trunk.port_id).sub_ports
        self.assertEqual(21, len(subs))
        self.assertEqual(new, subs[-1])

    def test_delete_trunk_subport(self):
        self.tapi.put_trunk(self.trunk.port_id, self.trunk)
        sub = self.trunk.sub_ports[10]
        self.tapi.delete_trunk_subport(self.trunk.id, sub)
        subs = self.tapi.get_trunk('ctx', self.trunk.port_id).sub_ports
        self.assertNotIn(sub, subs)
        self.assertEqual(19, len(subs))

    def test_get_trunk(self):
        self.tapi.put_trunk(self.trunk.port_id, self.trunk)
        self.assertEqual(self.trunk,
                         self.tapi.get_trunk('ctx', self.trunk.port_id))
        self.tapi.get_trunk('ctx', self.trunk.port_id)
        self.assertFalse(self.stub.get_trunk_details.called)

    def test_get_trunk_cache_miss(self):
        self.assertEqual(self.trunk,
                         self.tapi.get_trunk('ctx', self.trunk.port_id))
        self.tapi.get_trunk('ctx', self.trunk.port_id)
        self.assertEqual(1, len(self.stub.get_trunk_details.mock_calls))

    def test_get_trunk_not_found(self):
        self.stub.get_trunk_details.side_effect = (
            resources_rpc.ResourceNotFound(resource_id='1', resource_type='1'))
        self.assertIsNone(self.tapi.get_trunk('ctx', self.trunk.port_id))
        self.tapi.get_trunk('ctx', self.trunk.port_id)
        self.assertEqual(1, len(self.stub.get_trunk_details.mock_calls))

    def test_get_trunk_for_subport(self):
        self.tapi.put_trunk(self.trunk.port_id, self.trunk)
        t = self.tapi.get_trunk_for_subport(
            'ctx', self.trunk.sub_ports[0].port_id)
        self.assertEqual(self.trunk, t)



# Copyright (c) 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import uuid

import mock
from neutron_lib import constants as const
from oslo_config import cfg
from ovsdbapp.backend.ovs_idl import idlutils

from neutron.agent.common import ovs_lib
from neutron.agent.linux import ip_lib
from neutron.common import utils
from neutron.plugins.ml2.drivers.openvswitch.agent.common import (
    constants as agent_const)
from neutron.tests.common.exclusive_resources import port
from neutron.tests.common import net_helpers
from neutron.tests.functional.agent.linux import base


class OVSBridgeTestBase(base.BaseOVSLinuxTestCase):
    # TODO(twilson) So far, only ovsdb-related tests are written. It would be
    # good to also add the openflow-related functions
    def setUp(self):
        super(OVSBridgeTestBase, self).setUp()
        self.ovs = ovs_lib.BaseOVS()
        self.br = self.useFixture(net_helpers.OVSBridgeFixture()).bridge

    def create_ovs_port(self, *interface_attrs):
        # Convert ((a, b), (c, d)) to {a: b, c: d} and add 'type' by default
        attrs = collections.OrderedDict(interface_attrs)
        attrs.setdefault('type', 'internal')
        port_name = utils.get_rand_device_name(net_helpers.PORT_PREFIX)
        return (port_name, self.br.add_port(port_name, *attrs.items()))

    def create_ovs_vif_port(self, iface_id=None, mac=None,
                            iface_field='iface-id'):
        if iface_id is None:
            iface_id = utils.get_rand_name()
        if mac is None:
            mac = utils.get_rand_name()
        attrs = ('external_ids', {iface_field: iface_id, 'attached-mac': mac})
        port_name, ofport = self.create_ovs_port(attrs)
        return ovs_lib.VifPort(port_name, ofport, iface_id, mac, self.br)


class OVSBridgeTestCase(OVSBridgeTestBase):

    def test_port_lifecycle(self):
        (port_name, ofport) = self.create_ovs_port(('type', 'internal'))
        # ofport should always be an integer string with value -1 or > 0.
        self.assertTrue(int(ofport))
        self.assertTrue(int(self.br.get_port_ofport(port_name)))
        self.assertTrue(self.br.port_exists(port_name))
        self.assertEqual(self.br.br_name,
                         self.br.get_bridge_for_iface(port_name))
        self.br.delete_port(port_name)
        self.assertFalse(self.br.port_exists(port_name))

    def test_duplicate_port_may_exist_false(self):
        port_name, ofport = self.create_ovs_port(('type', 'internal'))
        cmd = self.br.ovsdb.add_port(self.br.br_name,
                                     port_name, may_exist=False)
        self.assertRaises(RuntimeError, cmd.execute, check_error=True)

    def test_delete_port_if_exists_false(self):
        cmd = self.br.ovsdb.del_port('nonexistantport', if_exists=False)
        self.assertRaises(RuntimeError, cmd.execute, check_error=True)

    def test_replace_port(self):
        port_name = utils.get_rand_device_name(net_helpers.PORT_PREFIX)
        self.br.replace_port(port_name, ('type', 'internal'))
        self.assertTrue(self.br.port_exists(port_name))
        self.assertEqual('internal',
                         self.br.db_get_val('Interface', port_name, 'type'))
        self.br.replace_port(port_name, ('type', 'internal'),
                             ('external_ids', {'test': 'test'}))
        self.assertTrue(self.br.port_exists(port_name))
        self.assertEqual('test', self.br.db_get_val('Interface', port_name,
                                                    'external_ids')['test'])
        self.assertEqual(agent_const.DEAD_VLAN_TAG,
                         self.br.db_get_val('Port', port_name, 'tag'))

    def test_attribute_lifecycle(self):
        (port_name, ofport) = self.create_ovs_port()
        tag = 42
        self.ovs.set_db_attribute('Port', port_name, 'tag', tag)
        self.assertEqual(tag, self.ovs.db_get_val('Port', port_name, 'tag'))
        self.assertEqual(tag, self.br.get_port_tag_dict()[port_name])
        self.ovs.clear_db_attribute('Port', port_name, 'tag')
        self.assertEqual([], self.ovs.db_get_val('Port', port_name, 'tag'))
        self.assertEqual([], self.br.get_port_tag_dict()[port_name])

    def test_attribute_map_handling(self):
        (pname, ofport) = self.create_ovs_port()
        expected = {'a': 'b'}
        self.ovs.set_db_attribute('Port', pname, 'other_config', expected)
        self.assertEqual(expected,
                         self.ovs.db_get_val('Port', pname, 'other_config'))
        other = {'c': 'd'}
        expected.update(other)
        self.ovs.set_db_attribute('Port', pname, 'other_config', other)
        self.assertEqual(expected,
                         self.ovs.db_get_val('Port', pname, 'other_config'))
        other = {'a': 'x'}
        expected.update(other)
        self.ovs.set_db_attribute('Port', pname, 'other_config', other)
        self.assertEqual(expected,
                         self.ovs.db_get_val('Port', pname, 'other_config'))

    def test_get_bridge_external_bridge_id(self):
        self.ovs.set_db_attribute('Bridge', self.br.br_name,
                                  'external_ids',
                                  {'bridge-id': self.br.br_name})
        self.assertEqual(
            self.br.br_name,
            self.ovs.get_bridge_external_bridge_id(self.br.br_name))

    def test_controller_lifecycle(self):
        controllers = {'tcp:127.0.0.1:6633', 'tcp:172.17.16.10:55'}
        self.br.set_controller(controllers)
        self.assertSetEqual(controllers, set(self.br.get_controller()))
        self.br.del_controller()
        self.assertEqual([], self.br.get_controller())

    def test_non_index_queries(self):
        controllers = ['tcp:127.0.0.1:6633']
        self.br.set_controller(controllers)
        cmd = self.br.ovsdb.db_set('Controller', self.br.br_name,
                                   ('connection_mode', 'out-of-band'))
        cmd.execute(check_error=True)
        self.assertEqual('out-of-band',
                         self.br.db_get_val('Controller', self.br.br_name,
                                            'connection_mode'))

    def test_set_fail_mode_secure(self):
        self.br.set_secure_mode()
        self._assert_br_fail_mode(ovs_lib.FAILMODE_SECURE)

    def test_set_fail_mode_standalone(self):
        self.br.set_standalone_mode()
        self._assert_br_fail_mode(ovs_lib.FAILMODE_STANDALONE)

    def _assert_br_fail_mode(self, fail_mode):
        self.assertEqual(
            self.br.db_get_val('Bridge', self.br.br_name, 'fail_mode'),
            fail_mode)

    def test_add_protocols_start_with_one(self):
        self.br.set_db_attribute('Bridge', self.br.br_name, 'protocols',
                                 ['OpenFlow10'],
                                 check_error=True)
        self.br.add_protocols('OpenFlow13')
        self.assertEqual(
            self.br.db_get_val('Bridge', self.br.br_name, 'protocols'),
            ['OpenFlow10', 'OpenFlow13'])

    def test_add_protocols_start_with_two_add_two(self):
        self.br.set_db_attribute('Bridge', self.br.br_name, 'protocols',
                                 ['OpenFlow10', 'OpenFlow12'],
                                 check_error=True)
        self.br.add_protocols('OpenFlow13', 'OpenFlow14')
        self.assertEqual(
            self.br.db_get_val('Bridge', self.br.br_name, 'protocols'),
            ['OpenFlow10', 'OpenFlow12', 'OpenFlow13', 'OpenFlow14'])

    def test_add_protocols_add_existing(self):
        self.br.set_db_attribute('Bridge', self.br.br_name, 'protocols',
                                 ['OpenFlow10', 'OpenFlow12', 'OpenFlow13'],
                                 check_error=True)
        self.br.add_protocols('OpenFlow13')
        self.assertEqual(
            self.br.db_get_val('Bridge', self.br.br_name, 'protocols'),
            ['OpenFlow10', 'OpenFlow12', 'OpenFlow13'])

    def test_get_datapath_id(self):
        brdev = ip_lib.IPDevice(self.br.br_name)
        dpid = brdev.link.attributes['link/ether'].replace(':', '')
        self.br.set_db_attribute('Bridge',
                                 self.br.br_name, 'datapath_id', dpid)
        self.assertIn(dpid, self.br.get_datapath_id())

    def _test_add_tunnel_port(self, attrs):
        port_name = utils.get_rand_device_name(net_helpers.PORT_PREFIX)
        self.br.add_tunnel_port(port_name, attrs['remote_ip'],
                                attrs['local_ip'])
        self.assertEqual('gre',
                         self.ovs.db_get_val('Interface', port_name, 'type'))
        options = self.ovs.db_get_val('Interface', port_name, 'options')
        for attr, val in attrs.items():
            self.assertEqual(val, options[attr])

    def test_add_tunnel_port_ipv4(self):
        attrs = {
            'remote_ip': self.get_test_net_address(1),
            'local_ip': self.get_test_net_address(2),
        }
        self._test_add_tunnel_port(attrs)

    def test_add_tunnel_port_ipv6(self):
        attrs = {
            'remote_ip': '2001:db8:200::1',
            'local_ip': '2001:db8:100::1',
        }
        self._test_add_tunnel_port(attrs)

    def test_add_tunnel_port_custom_port(self):
        port_name = utils.get_rand_device_name(net_helpers.PORT_PREFIX)
        self.br.add_tunnel_port(
            port_name,
            self.get_test_net_address(1),
            self.get_test_net_address(2),
            tunnel_type=const.TYPE_VXLAN,
            vxlan_udp_port=12345)
        options = self.ovs.db_get_val('Interface', port_name, 'options')
        self.assertEqual("12345", options['dst_port'])

    def test_add_tunnel_port_tos(self):
        attrs = {
            'remote_ip': self.get_test_net_address(1),
            'local_ip': self.get_test_net_address(2),
            'tos': 'inherit',
        }
        port_name = utils.get_rand_device_name(net_helpers.PORT_PREFIX)
        self.br.add_tunnel_port(port_name, attrs['remote_ip'],
                                attrs['local_ip'], tos=attrs['tos'])
        self.assertEqual('gre',
                         self.ovs.db_get_val('Interface', port_name, 'type'))
        options = self.ovs.db_get_val('Interface', port_name, 'options')
        for attr, val in attrs.items():
            self.assertEqual(val, options[attr])

    def test_add_patch_port(self):
        local = utils.get_rand_device_name(net_helpers.PORT_PREFIX)
        peer = 'remotepeer'
        self.br.add_patch_port(local, peer)
        self.assertEqual(self.ovs.db_get_val('Interface', local, 'type'),
                         'patch')
        options = self.ovs.db_get_val('Interface', local, 'options')
        self.assertEqual(peer, options['peer'])

    def test_get_port_name_list(self):
        # Note that ovs-vsctl's list-ports does not include the port created
        # with the same name as the bridge
        ports = {self.create_ovs_port()[0] for i in range(5)}
        self.assertSetEqual(ports, set(self.br.get_port_name_list()))

    def test_get_iface_name_list(self):
        ifaces = {self.create_ovs_port()[0] for i in range(5)}
        self.assertSetEqual(ifaces, set(self.br.get_iface_name_list()))

    def test_get_port_stats(self):
        # Nothing seems to use this function?
        (port_name, ofport) = self.create_ovs_port()
        stats = set(self.br.get_port_stats(port_name).keys())
        self.assertTrue(set(['rx_packets', 'tx_packets']).issubset(stats))

    def test_get_vif_ports(self):
        for i in range(2):
            self.create_ovs_port()
        vif_ports = [self.create_ovs_vif_port() for i in range(3)]
        ports = self.br.get_vif_ports()
        self.assertEqual(3, len(ports))
        self.assertTrue(all([isinstance(x, ovs_lib.VifPort) for x in ports]))
        self.assertEqual(sorted([x.port_name for x in vif_ports]),
                         sorted([x.port_name for x in ports]))

    def test_get_vif_ports_with_bond(self):
        for i in range(2):
            self.create_ovs_port()
        vif_ports = [self.create_ovs_vif_port() for i in range(3)]
        # bond ports don't have records in the Interface table but they do in
        # the Port table
        orig = self.br.get_port_name_list
        new_port_name_list = lambda: orig() + ['bondport']
        mock.patch.object(self.br, 'get_port_name_list',
                          new=new_port_name_list).start()
        ports = self.br.get_vif_ports()
        self.assertEqual(3, len(ports))
        self.assertTrue(all([isinstance(x, ovs_lib.VifPort) for x in ports]))
        self.assertEqual(sorted([x.port_name for x in vif_ports]),
                         sorted([x.port_name for x in ports]))

    def test_get_vif_port_set(self):
        for i in range(2):
            self.create_ovs_port()
        vif_ports = [self.create_ovs_vif_port() for i in range(2)]
        ports = self.br.get_vif_port_set()
        expected = set([x.vif_id for x in vif_ports])
        self.assertEqual(expected, ports)

    def test_get_vif_port_set_with_missing_port(self):
        self.create_ovs_port()
        vif_ports = [self.create_ovs_vif_port()]

        # return an extra port to make sure the db list ignores it
        orig = self.br.get_port_name_list
        new_port_name_list = lambda: orig() + ['anotherport']
        mock.patch.object(self.br, 'get_port_name_list',
                          new=new_port_name_list).start()
        ports = self.br.get_vif_port_set()
        expected = set([vif_ports[0].vif_id])
        self.assertEqual(expected, ports)

    def test_get_vif_port_set_on_empty_bridge_returns_empty_set(self):
        # Create a port on self.br
        self.create_ovs_vif_port()

        # Create another, empty bridge
        br_2 = self.useFixture(net_helpers.OVSBridgeFixture()).bridge

        # Assert that get_vif_port_set on an empty bridge returns an empty set,
        # and does not return the other bridge's ports.
        self.assertEqual(set(), br_2.get_vif_port_set())

    def test_get_ports_attributes(self):
        port_names = [self.create_ovs_port()[0], self.create_ovs_port()[0]]
        db_ports = self.br.get_ports_attributes('Interface', columns=['name'])
        db_ports_names = [p['name'] for p in db_ports]
        self.assertEqual(sorted(port_names), sorted(db_ports_names))

    def test_get_port_tag_dict(self):
        # Simple case tested in port test_set_get_clear_db_val
        pass

    def test_get_vif_port_by_id(self):
        for i in range(2):
            self.create_ovs_port()
        vif_ports = [self.create_ovs_vif_port() for i in range(3)]
        for vif in vif_ports:
            self.assertEqual(self.br.get_vif_port_by_id(vif.vif_id).vif_id,
                             vif.vif_id)

    def test_get_vifs_by_ids(self):
        for i in range(2):
            self.create_ovs_port()
        vif_ports = [self.create_ovs_vif_port() for i in range(3)]
        by_id = self.br.get_vifs_by_ids([v.vif_id for v in vif_ports])
        # convert to str for comparison of VifPorts
        by_id = {vid: str(vport) for vid, vport in by_id.items()}
        self.assertEqual({v.vif_id: str(v) for v in vif_ports}, by_id)

    def test_delete_ports(self):
        # TODO(twilson) I intensely dislike the current delete_ports function
        # as the default behavior is really delete_vif_ports(), then it acts
        # more like a delete_ports() seems like it should if all_ports=True is
        # passed
        # Create 2 non-vif ports and 2 vif ports
        nonvifs = {self.create_ovs_port()[0] for i in range(2)}
        vifs = {self.create_ovs_vif_port().port_name for i in range(2)}
        self.assertSetEqual(nonvifs.union(vifs),
                            set(self.br.get_port_name_list()))
        self.br.delete_ports()
        self.assertSetEqual(nonvifs, set(self.br.get_port_name_list()))
        self.br.delete_ports(all_ports=True)
        self.assertEqual(len(self.br.get_port_name_list()), 0)

    def test_set_controller_connection_mode(self):
        controllers = ['tcp:192.0.2.0:6633']
        self._set_controllers_connection_mode(controllers)

    def test_set_multi_controllers_connection_mode(self):
        controllers = ['tcp:192.0.2.0:6633', 'tcp:192.0.2.1:55']
        self._set_controllers_connection_mode(controllers)

    def _set_controllers_connection_mode(self, controllers):
        self.br.set_controller(controllers)
        self.assertEqual(sorted(controllers), sorted(self.br.get_controller()))
        self.br.set_controllers_connection_mode('out-of-band')
        self._assert_controllers_connection_mode('out-of-band')
        self.br.del_controller()
        self.assertEqual([], self.br.get_controller())

    def _assert_controllers_connection_mode(self, connection_mode):
        controllers = self.br.db_get_val('Bridge', self.br.br_name,
                                         'controller')
        controllers = [controllers] if isinstance(
            controllers, uuid.UUID) else controllers
        for controller in controllers:
            self.assertEqual(connection_mode,
                             self.br.db_get_val('Controller',
                                                controller,
                                                'connection_mode'))

    def test_egress_bw_limit(self):
        port_name, _ = self.create_ovs_port()
        self.br.create_egress_bw_limit_for_port(port_name, 700, 70)
        max_rate, burst = self.br.get_egress_bw_limit_for_port(port_name)
        self.assertEqual(700, max_rate)
        self.assertEqual(70, burst)
        self.br.delete_egress_bw_limit_for_port(port_name)
        max_rate, burst = self.br.get_egress_bw_limit_for_port(port_name)
        self.assertIsNone(max_rate)
        self.assertIsNone(burst)

    def test_ingress_bw_limit(self):
        port_name, _ = self.create_ovs_port()
        self.br.update_ingress_bw_limit_for_port(port_name, 700, 70)
        max_rate, burst = self.br.get_ingress_bw_limit_for_port(port_name)
        self.assertEqual(700, max_rate)
        self.assertEqual(70, burst)

        self.br.update_ingress_bw_limit_for_port(port_name, 750, 100)
        max_rate, burst = self.br.get_ingress_bw_limit_for_port(port_name)
        self.assertEqual(750, max_rate)
        self.assertEqual(100, burst)

        self.br.delete_ingress_bw_limit_for_port(port_name)
        max_rate, burst = self.br.get_ingress_bw_limit_for_port(port_name)
        self.assertIsNone(max_rate)
        self.assertIsNone(burst)

    def test_ingress_bw_limit_dpdk_port(self):
        port_name, _ = self.create_ovs_port(
            ('type', agent_const.OVS_DPDK_VHOST_USER))
        self.br.update_ingress_bw_limit_for_port(port_name, 700, 70)
        max_rate, burst = self.br.get_ingress_bw_limit_for_dpdk_port(
            port_name)
        self.assertEqual(700, max_rate)
        self.assertEqual(70, burst)

        self.br.update_ingress_bw_limit_for_port(port_name, 750, 100)
        max_rate, burst = self.br.get_ingress_bw_limit_for_dpdk_port(
            port_name)
        self.assertEqual(750, max_rate)
        self.assertEqual(100, burst)

        self.br.delete_ingress_bw_limit_for_port(port_name)
        max_rate, burst = self.br.get_ingress_bw_limit_for_dpdk_port(
            port_name)
        self.assertIsNone(max_rate)
        self.assertIsNone(burst)

    def test_db_create_references(self):
        with self.ovs.ovsdb.transaction(check_error=True) as txn:
            queue = txn.add(self.ovs.ovsdb.db_create("Queue",
                                                     other_config={'a': '1'}))
            qos = txn.add(self.ovs.ovsdb.db_create("QoS", queues={0: queue}))
            txn.add(self.ovs.ovsdb.db_set("Port", self.br.br_name,
                                          ('qos', qos)))

        def cleanup():
            with self.ovs.ovsdb.transaction() as t:
                t.add(self.ovs.ovsdb.db_destroy("QoS", qos.result))
                t.add(self.ovs.ovsdb.db_destroy("Queue", queue.result))
                t.add(self.ovs.ovsdb.db_clear("Port", self.br.br_name, 'qos'))

        self.addCleanup(cleanup)
        val = self.ovs.ovsdb.db_get("Port", self.br.br_name, 'qos').execute()
        self.assertEqual(qos.result, val)

    def test_db_add_set(self):
        protocols = ["OpenFlow10", "OpenFlow11"]
        self.br.ovsdb.db_add("Bridge", self.br.br_name, "protocols",
                             *protocols).execute(check_error=True)
        self.assertEqual(protocols,
                         self.br.db_get_val('Bridge',
                                            self.br.br_name, "protocols"))

    def test_db_add_map(self):
        key = "testdata"
        data = {key: "testvalue"}
        self.br.ovsdb.db_add("Bridge", self.br.br_name, "external_ids",
                             data).execute(check_error=True)
        self.assertEqual(data, self.br.db_get_val('Bridge', self.br.br_name,
                                                  'external_ids'))
        self.br.ovsdb.db_add("Bridge", self.br.br_name, "external_ids",
                             {key: "newdata"}).execute(check_error=True)
        self.assertEqual(data, self.br.db_get_val('Bridge', self.br.br_name,
                                                  'external_ids'))

    def test_db_add_map_multiple_one_dict(self):
        data = {"one": "1", "two": "2", "three": "3"}
        self.br.ovsdb.db_add("Bridge", self.br.br_name, "external_ids",
                             data).execute(check_error=True)
        self.assertEqual(data, self.br.db_get_val('Bridge', self.br.br_name,
                                                  'external_ids'))

    def test_db_add_map_multiple_dicts(self):
        data = ({"one": "1"}, {"two": "2"}, {"three": "3"})
        self.br.ovsdb.db_add("Bridge", self.br.br_name, "external_ids",
                             *data).execute(check_error=True)
        combined = {k: v for a in data for k, v in a.items()}
        self.assertEqual(combined, self.br.db_get_val('Bridge',
                                                      self.br.br_name,
                                                      'external_ids'))

    def test_db_add_ref(self):
        ovsdb = self.ovs.ovsdb
        brname = utils.get_rand_name(prefix=net_helpers.BR_PREFIX)
        br = ovs_lib.OVSBridge(brname)  # doesn't create
        self.addCleanup(br.destroy)
        with ovsdb.transaction(check_error=True) as txn:
            br = txn.add(ovsdb.db_create('Bridge', name=brname))
            txn.add(ovsdb.db_add('Open_vSwitch', '.', 'bridges', br))

        self.assertIn(brname, self.ovs.get_bridges())

    def test_db_add_to_new_object(self):
        ovsdb = self.ovs.ovsdb
        brname = utils.get_rand_name(prefix=net_helpers.BR_PREFIX)
        br = ovs_lib.OVSBridge(brname)  # doesn't create
        self.addCleanup(br.destroy)
        with ovsdb.transaction(check_error=True) as txn:
            txn.add(ovsdb.add_br(brname))
            txn.add(ovsdb.db_add('Bridge', brname, 'protocols', 'OpenFlow10'))

    def test_cascading_del_in_txn(self):
        ovsdb = self.ovs.ovsdb
        port_name, _ = self.create_ovs_port()

        def del_port_mod_iface():
            with ovsdb.transaction(check_error=True) as txn:
                txn.add(ovsdb.del_port(port_name, self.br.br_name,
                                       if_exists=False))
                txn.add(ovsdb.db_set('Interface', port_name,
                                     ('type', 'internal')))
        # native gives a more specific exception than vsctl
        self.assertRaises((RuntimeError, idlutils.RowNotFound),
                          del_port_mod_iface)

    def test_delete_flows_all(self):
        self.br.add_flow(in_port=1, actions="output:2")
        self.br.delete_flows(cookie=ovs_lib.COOKIE_ANY)
        self.assertEqual([], self.br.dump_all_flows())

    def test_delete_flows_strict(self):
        self.br.delete_flows(cookie=ovs_lib.COOKIE_ANY)  # remove NORMAL action
        self.br.add_flow(in_port=1, actions="output:2")
        self.br.add_flow(in_port=1, priority=100, actions="output:3")
        self.assertEqual(2, len(self.br.dump_all_flows()))
        self.br.delete_flows(in_port=1, priority=100, strict=True)
        self.assertEqual(1, len(self.br.dump_all_flows()))


class OVSLibTestCase(base.BaseOVSLinuxTestCase):

    def setUp(self):
        super(OVSLibTestCase, self).setUp()
        self.ovs = ovs_lib.BaseOVS()

    def test_add_manager_appends(self):
        port1 = self.useFixture(port.ExclusivePort(const.PROTO_NAME_TCP,
            start=net_helpers.OVS_MANAGER_TEST_PORT_FIRST,
            end=net_helpers.OVS_MANAGER_TEST_PORT_LAST)).port
        port2 = self.useFixture(port.ExclusivePort(const.PROTO_NAME_TCP,
            start=net_helpers.OVS_MANAGER_TEST_PORT_FIRST,
            end=net_helpers.OVS_MANAGER_TEST_PORT_LAST)).port
        manager_list = ["ptcp:%s:127.0.0.1" % port1,
                        "ptcp:%s:127.0.0.1" % port2]
        # Verify that add_manager does not override the existing manager
        expected_manager_list = list()
        for conn_uri in manager_list:
            self.ovs.add_manager(conn_uri)
            self.addCleanup(self.ovs.remove_manager, conn_uri)
            self.assertIn(conn_uri, self.ovs.get_manager())
            expected_manager_list.append(conn_uri)

        # Verify that switch is configured with both the managers
        for manager_uri in expected_manager_list:
            self.assertIn(manager_uri, manager_list)

    def test_add_manager_lifecycle_baseovs(self):
        port1 = self.useFixture(port.ExclusivePort(const.PROTO_NAME_TCP,
            start=net_helpers.OVS_MANAGER_TEST_PORT_FIRST,
            end=net_helpers.OVS_MANAGER_TEST_PORT_LAST)).port
        conn_uri = "ptcp:%s:127.0.0.1" % port1
        self.addCleanup(self.ovs.remove_manager, conn_uri)
        self.ovs.add_manager(conn_uri)
        self.assertIn(conn_uri, self.ovs.get_manager())
        self.assertEqual(self.ovs.db_get_val('Manager', conn_uri,
                                             'inactivity_probe'),
                         self.ovs.vsctl_timeout * 1000)
        self.ovs.remove_manager(conn_uri)
        self.assertNotIn(conn_uri, self.ovs.get_manager())

    def test_bridge_lifecycle_baseovs(self):
        name = utils.get_rand_name(prefix=net_helpers.BR_PREFIX)
        self.addCleanup(self.ovs.delete_bridge, name)
        br = self.ovs.add_bridge(name)
        self.assertEqual(br.br_name, name)
        self.assertTrue(self.ovs.bridge_exists(name))
        self.ovs.delete_bridge(name)
        self.assertFalse(self.ovs.bridge_exists(name))

    def test_get_bridges(self):
        bridges = {
            self.useFixture(net_helpers.OVSBridgeFixture()).bridge.br_name
            for i in range(5)}
        self.assertTrue(set(self.ovs.get_bridges()).issuperset(bridges))

    def test_bridge_lifecycle_ovsbridge(self):
        name = utils.get_rand_name(prefix=net_helpers.BR_PREFIX)
        mac_table_size = 12345
        cfg.CONF.set_override(
            'bridge_mac_table_size', mac_table_size, group='OVS')
        br = ovs_lib.OVSBridge(name)
        self.assertEqual(br.br_name, name)
        # Make sure that instantiating an OVSBridge does not actually create
        self.assertFalse(self.ovs.bridge_exists(name))
        self.addCleanup(self.ovs.delete_bridge, name)
        br.create()
        self.assertTrue(self.ovs.bridge_exists(name))
        br_other_config = self.ovs.ovsdb.db_find(
            'Bridge', ('name', '=', name), columns=['other_config']
        ).execute()[0]['other_config']
        self.assertEqual(str(mac_table_size),
                         br_other_config['mac-table-size'])
        br.destroy()
        self.assertFalse(self.ovs.bridge_exists(name))

    def test_db_find_column_type_list(self):
        """Fixate output for vsctl/native ovsdb_interface.

        Makes sure that db_find search queries give the same result for both
        implementations.
        """
        bridge_name = utils.get_rand_name(prefix=net_helpers.BR_PREFIX)
        self.addCleanup(self.ovs.delete_bridge, bridge_name)
        br = self.ovs.add_bridge(bridge_name)
        port_name = utils.get_rand_name(prefix=net_helpers.PORT_PREFIX)
        br.add_port(port_name)
        self.ovs.set_db_attribute('Port', port_name, 'tag', 42)

        # wrap list/find in transaction so we get a single isolated snapshot
        with self.ovs.ovsdb.transaction(check_error=True) as txn:
            tags = txn.add(self.ovs.ovsdb.db_list('Port', columns=['tag']))
            len_0_list = txn.add(self.ovs.ovsdb.db_find(
                    'Port', ('tag', '!=', []), columns=['tag']))
            single_value = txn.add(self.ovs.ovsdb.db_find(
                    'Port', ('tag', '=', 42), columns=['tag']))

        # Make sure that there is data to query.
        # It should be, but let's be a little paranoid here as otherwise
        # the test has no sense
        tags_present = [t for t in tags.result if t['tag'] != []]
        self.assertTrue(tags_present)
        tags_42 = [t for t in tags_present if t['tag'] == 42]
        self.assertEqual(tags_42, single_value.result)
        self.assertItemsEqual(len_0_list.result, tags_present)



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
from operator import attrgetter
import random

import six


@six.add_metaclass(abc.ABCMeta)
class BaseScheduler(object):
    """The base scheduler (agnostic to resource type).
       Child classes of BaseScheduler must define the
       self.resource_filter to filter agents of
       particular type.
    """
    resource_filter = None

    @abc.abstractmethod
    def select(self, plugin, context, resource_hostable_agents,
               resource_hosted_agents, num_agents_needed):
        """Return a subset of agents based on the specific scheduling logic."""

    def schedule(self, plugin, context, resource):
        """Select and bind agents to a given resource."""
        if not self.resource_filter:
            return
        # filter the agents that can host the resource
        filtered_agents_dict = self.resource_filter.filter_agents(
            plugin, context, resource)
        num_agents = filtered_agents_dict['n_agents']
        hostable_agents = filtered_agents_dict['hostable_agents']
        hosted_agents = filtered_agents_dict['hosted_agents']
        chosen_agents = self.select(plugin, context, hostable_agents,
                                    hosted_agents, num_agents)
        # bind the resource to the agents
        self.resource_filter.bind(context, chosen_agents, resource['id'])
        return chosen_agents


class BaseChanceScheduler(BaseScheduler):
    """Choose agents randomly."""

    def __init__(self, resource_filter):
        self.resource_filter = resource_filter

    def select(self, plugin, context, resource_hostable_agents,
               resource_hosted_agents, num_agents_needed):
        chosen_agents = random.sample(resource_hostable_agents,
                                      num_agents_needed)
        return chosen_agents


class BaseWeightScheduler(BaseScheduler):
    """Choose agents based on load."""

    def __init__(self, resource_filter):
        self.resource_filter = resource_filter

    def select(self, plugin, context, resource_hostable_agents,
               resource_hosted_agents, num_agents_needed):
        chosen_agents = sorted(resource_hostable_agents,
                           key=attrgetter('load'))[0:num_agents_needed]
        return chosen_agents



# Copyright 2015 HuaWei Technologies.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_utils import timeutils
from sqlalchemy.orm import session as se

from neutron.db import _model_query as model_query
from neutron.db import _resource_extend as resource_extend
from neutron.db import api as db_api
from neutron.db import standard_attr

CHANGED_SINCE = 'changed_since'
TIME_FORMAT_WHOLE_SECONDS = '%Y-%m-%dT%H:%M:%S'


def _change_since_result_filter_hook(query, filters):
    # this block is for change_since query
    # we get the changed_since string from filters.
    # And translate it from string to datetime type.
    # Then compare with the timestamp in db which has
    # datetime type.
    values = filters and filters.get(CHANGED_SINCE, [])
    if not values:
        return query
    data = filters[CHANGED_SINCE][0]
    try:
        changed_since_string = timeutils.parse_isotime(data)
    except Exception:
        msg = ("The input %s must be in the "
               "following format: YYYY-MM-DDTHH:MM:SSZ") % CHANGED_SINCE
        raise n_exc.InvalidInput(error_message=msg)
    changed_since = (timeutils.
                     normalize_time(changed_since_string))
    target_model_class = query.column_descriptions[0]['type']
    query = query.join(standard_attr.StandardAttribute,
                       target_model_class.standard_attr_id ==
                       standard_attr.StandardAttribute.id).filter(
                       standard_attr.StandardAttribute.updated_at
                       >= changed_since)
    return query


def _update_timestamp(session, context, instances):
    objs_list = session.new.union(session.dirty)

    while objs_list:
        obj = objs_list.pop()
        if (isinstance(obj, standard_attr.HasStandardAttributes)
            and obj.standard_attr_id):
            obj.updated_at = timeutils.utcnow()


def _format_timestamp(resource_db, result):
    result['created_at'] = (resource_db.created_at.
                            strftime(TIME_FORMAT_WHOLE_SECONDS)) + 'Z'
    result['updated_at'] = (resource_db.updated_at.
                            strftime(TIME_FORMAT_WHOLE_SECONDS)) + 'Z'


def _add_timestamp(mapper, _conn, target):
    if not target.created_at and not target.updated_at:
        time = timeutils.utcnow()
        for field in ['created_at', 'updated_at']:
            setattr(target, field, time)
    return target


@resource_extend.has_resource_extenders
class TimeStamp_db_mixin(object):
    """Mixin class to add Time Stamp methods."""

    def __new__(cls, *args, **kwargs):
        rs_model_maps = standard_attr.get_standard_attr_resource_model_map()
        for model in rs_model_maps.values():
            model_query.register_hook(
                model,
                "change_since_query",
                query_hook=None,
                filter_hook=None,
                result_filters=_change_since_result_filter_hook)
        return super(TimeStamp_db_mixin, cls).__new__(cls, *args, **kwargs)

    def register_db_events(self):
        listen = db_api.sqla_listen
        listen(standard_attr.StandardAttribute, 'before_insert',
               _add_timestamp)
        listen(se.Session, 'before_flush', _update_timestamp)

    @staticmethod
    @resource_extend.extends(
        list(standard_attr.get_standard_attr_resource_model_map()))
    def _extend_resource_dict_timestamp(resource_res, resource_db):
        if (resource_db and resource_db.created_at and
                resource_db.updated_at):
            _format_timestamp(resource_db, resource_res)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six

from neutron.agent import agent_extension


@six.add_metaclass(abc.ABCMeta)
class L2AgentExtension(agent_extension.AgentExtension):
    """Define stable abstract interface for l2 agent extensions.

    An agent extension extends the agent core functionality.
    """

    def initialize(self, connection, driver_type):
        """Initialize agent extension."""

    @abc.abstractmethod
    def handle_port(self, context, data):
        """Handle agent extension for port.

        This can be called on either create or update, depending on the
        code flow. Thus, it's this function's responsibility to check what
        actually changed.

        :param context: rpc context
        :param data: port data
        """

    @abc.abstractmethod
    def delete_port(self, context, data):
        """Delete port from agent extension.

        :param context: rpc context
        :param data: port data
        """



# Copyright (c) 2015 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from neutron_lib import context
from oslo_utils import uuidutils
from oslo_versionedobjects import fields as obj_fields
import testtools

from neutron.api.rpc.callbacks import resources
from neutron.api.rpc.callbacks import version_manager
from neutron.api.rpc.handlers import resources_rpc
from neutron.common import topics
from neutron.objects import base as objects_base
from neutron.objects import common_types
from neutron.tests import base
from neutron.tests.unit.objects import test_base as objects_test_base


TEST_EVENT = 'test_event'
TEST_VERSION = '1.0'


def _create_test_dict(uuid=None):
    return {'id': uuid or uuidutils.generate_uuid(),
            'field': 'foo'}


def _create_test_resource(context=None, resource_cls=None):
    resource_cls = resource_cls or FakeResource
    resource_dict = _create_test_dict()
    resource = resource_cls(context, **resource_dict)
    resource.obj_reset_changes()
    return resource


class BaseFakeResource(objects_base.NeutronObject):
    @classmethod
    def get_objects(cls, context, **kwargs):
        return list()


class FakeResource(BaseFakeResource):
    VERSION = TEST_VERSION

    fields = {
        'id': common_types.UUIDField(),
        'field': obj_fields.StringField()
    }


class FakeResource2(BaseFakeResource):
    VERSION = TEST_VERSION

    fields = {
        'id': common_types.UUIDField(),
        'field': obj_fields.StringField()
    }


class ResourcesRpcBaseTestCase(base.BaseTestCase):

    def setUp(self):
        super(ResourcesRpcBaseTestCase, self).setUp()

        self.obj_registry = self.useFixture(
            objects_test_base.NeutronObjectRegistryFixture())

        self.context = context.get_admin_context()
        mock.patch.object(resources_rpc.resources,
                          'is_valid_resource_type').start()
        mock.patch.object(resources_rpc.resources, 'get_resource_cls',
                          side_effect=self._get_resource_cls).start()

        self.resource_objs = [_create_test_resource(self.context)
                              for _ in range(2)]
        self.resource_objs2 = [_create_test_resource(self.context,
                                                     FakeResource2)
                               for _ in range(2)]

    @staticmethod
    def _get_resource_cls(resource_type):
        return {FakeResource.obj_name(): FakeResource,
                FakeResource2.obj_name(): FakeResource2}.get(resource_type)


class _ValidateResourceTypeTestCase(base.BaseTestCase):
    def setUp(self):
        super(_ValidateResourceTypeTestCase, self).setUp()
        self.is_valid_mock = mock.patch.object(
            resources_rpc.resources, 'is_valid_resource_type').start()

    def test_valid_type(self):
        self.is_valid_mock.return_value = True
        resources_rpc._validate_resource_type('foo')

    def test_invalid_type(self):
        self.is_valid_mock.return_value = False
        with testtools.ExpectedException(
                resources_rpc.InvalidResourceTypeClass):
            resources_rpc._validate_resource_type('foo')


class _ResourceTypeVersionedTopicTestCase(base.BaseTestCase):

    @mock.patch.object(resources_rpc, '_validate_resource_type')
    def test_resource_type_versioned_topic(self, validate_mock):
        obj_name = FakeResource.obj_name()
        expected = topics.RESOURCE_TOPIC_PATTERN % {
            'resource_type': 'FakeResource', 'version': '1.0'}
        with mock.patch.object(resources_rpc.resources, 'get_resource_cls',
                return_value=FakeResource):
            observed = resources_rpc.resource_type_versioned_topic(obj_name)
        self.assertEqual(expected, observed)


class ResourcesPullRpcApiTestCase(ResourcesRpcBaseTestCase):

    def setUp(self):
        super(ResourcesPullRpcApiTestCase, self).setUp()
        self.rpc = resources_rpc.ResourcesPullRpcApi()
        mock.patch.object(self.rpc, 'client').start()
        self.cctxt_mock = self.rpc.client.prepare.return_value

    def test_is_singleton(self):
        self.assertIs(self.rpc, resources_rpc.ResourcesPullRpcApi())

    def test_pull(self):
        self.obj_registry.register(FakeResource)
        expected_obj = _create_test_resource(self.context)
        resource_id = expected_obj.id
        self.cctxt_mock.call.return_value = expected_obj.obj_to_primitive()

        result = self.rpc.pull(
            self.context, FakeResource.obj_name(), resource_id)

        self.cctxt_mock.call.assert_called_once_with(
            self.context, 'pull', resource_type='FakeResource',
            version=TEST_VERSION, resource_id=resource_id)
        self.assertEqual(expected_obj, result)

    def test_bulk_pull(self):
        self.obj_registry.register(FakeResource)
        expected_objs = [_create_test_resource(self.context),
                         _create_test_resource(self.context)]
        self.cctxt_mock.call.return_value = [
            e.obj_to_primitive() for e in expected_objs]

        filter_kwargs = {'a': 'b', 'c': 'd'}
        result = self.rpc.bulk_pull(
            self.context, FakeResource.obj_name(),
            filter_kwargs=filter_kwargs)

        self.cctxt_mock.call.assert_called_once_with(
            self.context, 'bulk_pull', resource_type='FakeResource',
            version=TEST_VERSION, filter_kwargs=filter_kwargs)
        self.assertEqual(expected_objs, result)

    def test_pull_resource_not_found(self):
        resource_dict = _create_test_dict()
        resource_id = resource_dict['id']
        self.cctxt_mock.call.return_value = None
        with testtools.ExpectedException(resources_rpc.ResourceNotFound):
            self.rpc.pull(self.context, FakeResource.obj_name(),
                          resource_id)


class ResourcesPushToServerRpcCallbackTestCase(ResourcesRpcBaseTestCase):

    def test_report_versions(self):
        callbacks = resources_rpc.ResourcesPushToServerRpcCallback()
        with mock.patch('neutron.api.rpc.callbacks.version_manager'
                        '.update_versions') as update_versions:
            version_map = {'A': '1.0'}
            callbacks.report_agent_resource_versions(context=mock.ANY,
                                      agent_type='DHCP Agent',
                                      agent_host='fake-host',
                                      version_map=version_map)
            update_versions.assert_called_once_with(mock.ANY,
                                                    version_map)


class ResourcesPullRpcCallbackTestCase(ResourcesRpcBaseTestCase):

    def setUp(self):
        super(ResourcesPullRpcCallbackTestCase, self).setUp()
        self.obj_registry.register(FakeResource)
        self.callbacks = resources_rpc.ResourcesPullRpcCallback()
        self.resource_obj = _create_test_resource(self.context)

    def test_pull(self):
        resource_dict = _create_test_dict(uuid=self.resource_obj.id)
        with mock.patch.object(
                resources_rpc.prod_registry, 'pull',
                return_value=self.resource_obj) as registry_mock:
            primitive = self.callbacks.pull(
                self.context, resource_type=FakeResource.obj_name(),
                version=TEST_VERSION,
                resource_id=self.resource_obj.id)
        registry_mock.assert_called_once_with(
            'FakeResource', self.resource_obj.id, context=self.context)
        self.assertEqual(resource_dict,
                         primitive['versioned_object.data'])
        self.assertEqual(self.resource_obj.obj_to_primitive(), primitive)

    def test_bulk_pull(self):
        r1 = self.resource_obj
        r2 = _create_test_resource(self.context)

        @classmethod
        def get_objs(*args, **kwargs):
            if 'id' not in kwargs:
                return [r1, r2]
            return [r for r in [r1, r2] if r.id == kwargs['id']]

        # the bulk interface currently retrieves directly from the registry
        with mock.patch.object(FakeResource, 'get_objects', new=get_objs):
            objs = self.callbacks.bulk_pull(
                self.context, resource_type=FakeResource.obj_name(),
                version=TEST_VERSION)
            self.assertItemsEqual([r1.obj_to_primitive(),
                                   r2.obj_to_primitive()],
                                  objs)
            objs = self.callbacks.bulk_pull(
                self.context, resource_type=FakeResource.obj_name(),
                version=TEST_VERSION, filter_kwargs={'id': r1.id})
            self.assertEqual([r1.obj_to_primitive()], objs)

    @mock.patch.object(FakeResource, 'obj_to_primitive')
    def test_pull_backports_to_older_version(self, to_prim_mock):
        with mock.patch.object(resources_rpc.prod_registry, 'pull',
                               return_value=self.resource_obj):
            self.callbacks.pull(
                self.context, resource_type=FakeResource.obj_name(),
                version='0.9',  # less than initial version 1.0
                resource_id=self.resource_obj.id)
            to_prim_mock.assert_called_with(target_version='0.9')


class ResourcesPushRpcApiTestCase(ResourcesRpcBaseTestCase):
    """Tests the neutron server side of the RPC interface."""

    def setUp(self):
        super(ResourcesPushRpcApiTestCase, self).setUp()
        mock.patch.object(resources_rpc.n_rpc, 'get_client').start()
        self.rpc = resources_rpc.ResourcesPushRpcApi()
        self.cctxt_mock = self.rpc.client.prepare.return_value
        mock.patch.object(version_manager, 'get_resource_versions',
                         return_value=set([TEST_VERSION])).start()

    def test__prepare_object_fanout_context(self):
        expected_topic = topics.RESOURCE_TOPIC_PATTERN % {
            'resource_type': resources.get_resource_type(
                self.resource_objs[0]),
            'version': TEST_VERSION}

        observed = self.rpc._prepare_object_fanout_context(
            self.resource_objs[0], self.resource_objs[0].VERSION, '1.0')

        self.rpc.client.prepare.assert_called_once_with(
            fanout=True, topic=expected_topic, version='1.0')
        self.assertEqual(self.cctxt_mock, observed)

    def test_push_single_type(self):
        self.rpc.push(
            self.context, self.resource_objs, TEST_EVENT)

        self.cctxt_mock.cast.assert_called_once_with(
            self.context, 'push',
            resource_list=[resource.obj_to_primitive()
                           for resource in self.resource_objs],
            event_type=TEST_EVENT)

    def test_push_mixed(self):
        self.rpc.push(
            self.context, self.resource_objs + self.resource_objs2,
            event_type=TEST_EVENT)

        self.cctxt_mock.cast.assert_any_call(
            self.context, 'push',
            resource_list=[resource.obj_to_primitive()
                           for resource in self.resource_objs],
            event_type=TEST_EVENT)

        self.cctxt_mock.cast.assert_any_call(
            self.context, 'push',
            resource_list=[resource.obj_to_primitive()
                           for resource in self.resource_objs2],
            event_type=TEST_EVENT)


class ResourcesPushRpcCallbackTestCase(ResourcesRpcBaseTestCase):
    """Tests the agent-side of the RPC interface."""

    def setUp(self):
        super(ResourcesPushRpcCallbackTestCase, self).setUp()
        self.callbacks = resources_rpc.ResourcesPushRpcCallback()

    @mock.patch.object(resources_rpc.cons_registry, 'push')
    def test_push(self, reg_push_mock):
        self.obj_registry.register(FakeResource)
        self.callbacks.push(self.context,
                            resource_list=[resource.obj_to_primitive()
                                           for resource in self.resource_objs],
                            event_type=TEST_EVENT)
        reg_push_mock.assert_called_once_with(self.context,
                                              self.resource_objs[0].obj_name(),
                                              self.resource_objs,
                                              TEST_EVENT)



# Copyright (c) 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import contextlib
import sys
import time

from neutron_lib.agent import constants as agent_consts
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources as local_resources
from neutron_lib import constants
from neutron_lib import context
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_service import service
from oslo_utils import excutils
from osprofiler import profiler

from neutron.agent.l2 import l2_agent_extensions_manager as ext_manager
from neutron.agent import rpc as agent_rpc
from neutron.agent import securitygroups_rpc as agent_sg_rpc
from neutron.api.rpc.callbacks import resources
from neutron.api.rpc.handlers import securitygroups_rpc as sg_rpc
from neutron.common import config as common_config
from neutron.common import topics
from neutron.plugins.ml2.drivers.agent import _agent_manager_base as amb
from neutron.plugins.ml2.drivers.agent import capabilities
from neutron.plugins.ml2.drivers.agent import config as cagt_config  # noqa

LOG = logging.getLogger(__name__)


@profiler.trace_cls("rpc")
class CommonAgentLoop(service.Service):

    def __init__(self, manager, polling_interval,
                 quitting_rpc_timeout, agent_type, agent_binary):
        """Constructor.

        :param manager: the manager object containing the impl specifics
        :param polling_interval: interval (secs) to poll DB.
        :param quitting_rpc_timeout: timeout in seconds for rpc calls after
               stop is called.
        :param agent_type: Specifies the type of the agent
        :param agent_binary: The agent binary string
        """
        super(CommonAgentLoop, self).__init__()
        self.mgr = manager
        self._validate_manager_class()
        self.polling_interval = polling_interval
        self.quitting_rpc_timeout = quitting_rpc_timeout
        self.agent_type = agent_type
        self.agent_binary = agent_binary

    def _validate_manager_class(self):
        if not isinstance(self.mgr,
                          amb.CommonAgentManagerBase):
            LOG.error("Manager class must inherit from "
                      "CommonAgentManagerBase to ensure CommonAgent "
                      "works properly.")
            sys.exit(1)

    def start(self):
        # stores all configured ports on agent
        self.network_ports = collections.defaultdict(list)
        # flag to do a sync after revival
        self.fullsync = False
        self.context = context.get_admin_context_without_session()
        self.setup_rpc()
        self.init_extension_manager(self.connection)

        configurations = {'extensions': self.ext_manager.names()}
        configurations.update(self.mgr.get_agent_configurations())

        #TODO(mangelajo): optimize resource_versions (see ovs agent)
        self.agent_state = {
            'binary': self.agent_binary,
            'host': cfg.CONF.host,
            'topic': constants.L2_AGENT_TOPIC,
            'configurations': configurations,
            'agent_type': self.agent_type,
            'resource_versions': resources.LOCAL_RESOURCE_VERSIONS,
            'start_flag': True}

        report_interval = cfg.CONF.AGENT.report_interval
        if report_interval:
            heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            heartbeat.start(interval=report_interval)

        capabilities.notify_init_event(self.agent_type, self)
        # The initialization is complete; we can start receiving messages
        self.connection.consume_in_threads()

        self.daemon_loop()

    def stop(self, graceful=True):
        LOG.info("Stopping %s agent.", self.agent_type)
        if graceful and self.quitting_rpc_timeout:
            self.set_rpc_timeout(self.quitting_rpc_timeout)
        super(CommonAgentLoop, self).stop(graceful)

    def reset(self):
        common_config.setup_logging()

    def _report_state(self):
        try:
            devices = len(self.mgr.get_all_devices())
            self.agent_state.get('configurations')['devices'] = devices
            agent_status = self.state_rpc.report_state(self.context,
                                                       self.agent_state,
                                                       True)
            if agent_status == agent_consts.AGENT_REVIVED:
                LOG.info('%s Agent has just been revived. '
                         'Doing a full sync.',
                         self.agent_type)
                self.fullsync = True
            # we only want to update resource versions on startup
            self.agent_state.pop('resource_versions', None)
            self.agent_state.pop('start_flag', None)
        except Exception:
            LOG.exception("Failed reporting state!")

    def _validate_rpc_endpoints(self):
        if not isinstance(self.endpoints[0],
                          amb.CommonAgentManagerRpcCallBackBase):
            LOG.error("RPC Callback class must inherit from "
                      "CommonAgentManagerRpcCallBackBase to ensure "
                      "CommonAgent works properly.")
            sys.exit(1)

    def setup_rpc(self):
        self.plugin_rpc = agent_rpc.PluginApi(topics.PLUGIN)
        self.sg_plugin_rpc = sg_rpc.SecurityGroupServerRpcApi(topics.PLUGIN)
        self.sg_agent = agent_sg_rpc.SecurityGroupAgentRpc(
            self.context, self.sg_plugin_rpc, defer_refresh_firewall=True)

        self.agent_id = self.mgr.get_agent_id()
        LOG.info("RPC agent_id: %s", self.agent_id)

        self.topic = topics.AGENT
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.REPORTS)
        # RPC network init
        # Handle updates from service
        self.rpc_callbacks = self.mgr.get_rpc_callbacks(self.context, self,
                                                        self.sg_agent)
        self.endpoints = [self.rpc_callbacks]
        self._validate_rpc_endpoints()
        # Define the listening consumers for the agent
        consumers = self.mgr.get_rpc_consumers()
        self.connection = agent_rpc.create_consumers(self.endpoints,
                                                     self.topic,
                                                     consumers,
                                                     start_listening=False)

    def init_extension_manager(self, connection):
        ext_manager.register_opts(cfg.CONF)
        self.ext_manager = (
            ext_manager.L2AgentExtensionsManager(cfg.CONF))
        agent_api = self.mgr.get_agent_api(sg_agent=self.sg_agent)
        self.ext_manager.initialize(
            connection, self.mgr.get_extension_driver_type(), agent_api)

    def _clean_network_ports(self, device):
        for netid, ports_list in self.network_ports.items():
            for port_data in ports_list:
                if device == port_data['device']:
                    ports_list.remove(port_data)
                    if ports_list == []:
                        self.network_ports.pop(netid)
                    return port_data['port_id']

    def _update_network_ports(self, network_id, port_id, device):
        self._clean_network_ports(device)
        self.network_ports[network_id].append({
            "port_id": port_id,
            "device": device
        })

    def process_network_devices(self, device_info):
        resync_a = False
        resync_b = False

        self.sg_agent.setup_port_filters(device_info.get('added'),
                                         device_info.get('updated'))
        # Updated devices are processed the same as new ones, as their
        # admin_state_up may have changed. The set union prevents duplicating
        # work when a device is new and updated in the same polling iteration.
        devices_added_updated = (set(device_info.get('added'))
                                 | set(device_info.get('updated')))
        if devices_added_updated:
            resync_a = self.treat_devices_added_updated(devices_added_updated)

        if device_info.get('removed'):
            resync_b = self.treat_devices_removed(device_info['removed'])
        # If one of the above operations fails => resync with plugin
        return (resync_a | resync_b)

    def treat_devices_added_updated(self, devices):
        try:
            devices_details_list = self.plugin_rpc.get_devices_details_list(
                self.context, devices, self.agent_id, host=cfg.CONF.host)
        except Exception:
            LOG.exception("Unable to get port details for %s", devices)
            # resync is needed
            return True

        for device_details in devices_details_list:
            self._process_device_if_exists(device_details)
        # no resync is needed
        return False

    def _process_device_if_exists(self, device_details):
        # ignore exceptions from devices that disappear because they will
        # be handled as removed in the next iteration
        device = device_details['device']
        with self._ignore_missing_device_exceptions(device):
            LOG.debug("Port %s added", device)

            if 'port_id' in device_details:
                LOG.info("Port %(device)s updated. Details: %(details)s",
                         {'device': device, 'details': device_details})
                self.mgr.setup_arp_spoofing_protection(device,
                                                       device_details)

                segment = amb.NetworkSegment(
                    device_details.get('network_type'),
                    device_details['physical_network'],
                    device_details.get('segmentation_id'),
                    device_details.get('mtu')
                )
                network_id = device_details['network_id']
                self.rpc_callbacks.add_network(network_id, segment)
                interface_plugged = self.mgr.plug_interface(
                    network_id, segment,
                    device, device_details['device_owner'])
                # REVISIT(scheuran): Changed the way how ports admin_state_up
                # is implemented.
                #
                # Old lb implementation:
                # - admin_state_up: ensure that tap is plugged into bridge
                # - admin_state_down: remove tap from bridge
                # New lb implementation:
                # - admin_state_up: set tap device state to up
                # - admin_state_down: set tap device state to down
                #
                # However both approaches could result in races with
                # nova/libvirt and therefore to an invalid system state in the
                # scenario, where an instance is booted with a port configured
                # with admin_state_up = False:
                #
                # Libvirt does the following actions in exactly
                # this order (see libvirt virnetdevtap.c)
                #     1) Create the tap device, set its MAC and MTU
                #     2) Plug the tap into the bridge
                #     3) Set the tap online
                #
                # Old lb implementation:
                #   A race could occur, if the lb agent removes the tap device
                #   right after step 1). Then libvirt will add it to the bridge
                #   again in step 2).
                # New lb implementation:
                #   The race could occur if the lb-agent sets the taps device
                #   state to down right after step 2). In step 3) libvirt
                #   might set it to up again.
                #
                # This is not an issue if an instance is booted with a port
                # configured with admin_state_up = True. Libvirt would just
                # set the tap device up again.
                #
                # This refactoring is recommended for the following reasons:
                # 1) An existing race with libvirt caused by the behavior of
                #    the old implementation. See Bug #1312016
                # 2) The new code is much more readable
                if interface_plugged:
                    self.mgr.ensure_port_admin_state(
                        device, device_details['admin_state_up'])
                # update plugin about port status if admin_state is up
                if device_details['admin_state_up']:
                    if interface_plugged:
                        self.plugin_rpc.update_device_up(self.context,
                                                         device,
                                                         self.agent_id,
                                                         cfg.CONF.host)
                    else:
                        self.plugin_rpc.update_device_down(self.context,
                                                           device,
                                                           self.agent_id,
                                                           cfg.CONF.host)
                self._update_network_ports(device_details['network_id'],
                                           device_details['port_id'],
                                           device_details['device'])
                self.ext_manager.handle_port(self.context, device_details)
                registry.notify(local_resources.PORT_DEVICE,
                                events.AFTER_UPDATE, self,
                                context=self.context,
                                device_details=device_details)
            else:
                LOG.info("Device %s not defined on plugin", device)

    @contextlib.contextmanager
    def _ignore_missing_device_exceptions(self, device):
        try:
            yield
        except Exception:
            with excutils.save_and_reraise_exception() as ectx:
                if device not in self.mgr.get_all_devices():
                    ectx.reraise = False
                    LOG.debug("%s was removed during processing.", device)

    def treat_devices_removed(self, devices):
        resync = False
        self.sg_agent.remove_devices_filter(devices)
        for device in devices:
            LOG.info("Attachment %s removed", device)
            details = None
            try:
                details = self.plugin_rpc.update_device_down(self.context,
                                                             device,
                                                             self.agent_id,
                                                             cfg.CONF.host)
            except Exception:
                LOG.exception("Error occurred while removing port %s",
                              device)
                resync = True
            if details and details['exists']:
                LOG.info("Port %s updated.", device)
            else:
                LOG.debug("Device %s not defined on plugin", device)
            port_id = self._clean_network_ports(device)
            try:
                self.ext_manager.delete_port(self.context,
                                             {'device': device,
                                              'port_id': port_id})
            except Exception:
                LOG.exception("Error occurred while processing extensions "
                              "for port removal %s", device)
                resync = True
            registry.notify(local_resources.PORT_DEVICE, events.AFTER_DELETE,
                            self, context=self.context, device=device,
                            port_id=port_id)
        self.mgr.delete_arp_spoofing_protection(devices)
        return resync

    @staticmethod
    def _get_devices_locally_modified(timestamps, previous_timestamps):
        """Returns devices with previous timestamps that do not match new.

        If a device did not have a timestamp previously, it will not be
        returned because this means it is new.
        """
        return {device for device, timestamp in timestamps.items()
                if previous_timestamps.get(device) and
                timestamp != previous_timestamps.get(device)}

    def scan_devices(self, previous, sync):
        device_info = {}

        updated_devices = self.rpc_callbacks.get_and_clear_updated_devices()

        current_devices = self.mgr.get_all_devices()
        device_info['current'] = current_devices

        if previous is None:
            # This is the first iteration of daemon_loop().
            previous = {'added': set(),
                        'current': set(),
                        'updated': set(),
                        'removed': set(),
                        'timestamps': {}}
            # clear any orphaned ARP spoofing rules (e.g. interface was
            # manually deleted)
            self.mgr.delete_unreferenced_arp_protection(current_devices)

        # check to see if any devices were locally modified based on their
        # timestamps changing since the previous iteration. If a timestamp
        # doesn't exist for a device, this calculation is skipped for that
        # device.
        device_info['timestamps'] = self.mgr.get_devices_modified_timestamps(
            current_devices)
        locally_updated = self._get_devices_locally_modified(
            device_info['timestamps'], previous['timestamps'])
        if locally_updated:
            LOG.debug("Adding locally changed devices to updated set: %s",
                      locally_updated)
            updated_devices |= locally_updated

        if sync:
            # This is the first iteration, or the previous one had a problem.
            # Re-add all existing devices.
            device_info['added'] = current_devices

            # Retry cleaning devices that may not have been cleaned properly.
            # And clean any that disappeared since the previous iteration.
            device_info['removed'] = (previous['removed'] | previous['current']
                                      - current_devices)

            # Retry updating devices that may not have been updated properly.
            # And any that were updated since the previous iteration.
            # Only update devices that currently exist.
            device_info['updated'] = (previous['updated'] | updated_devices
                                      & current_devices)
        else:
            device_info['added'] = current_devices - previous['current']
            device_info['removed'] = previous['current'] - current_devices
            device_info['updated'] = updated_devices & current_devices

        return device_info

    def _device_info_has_changes(self, device_info):
        return (device_info.get('added')
                or device_info.get('updated')
                or device_info.get('removed'))

    def daemon_loop(self):
        LOG.info("%s Agent RPC Daemon Started!", self.agent_type)
        device_info = None
        sync = True

        while True:
            start = time.time()

            if self.fullsync:
                sync = True
                self.fullsync = False

            if sync:
                LOG.info("%s Agent out of sync with plugin!",
                         self.agent_type)

            device_info = self.scan_devices(previous=device_info, sync=sync)
            sync = False

            if (self._device_info_has_changes(device_info)
                or self.sg_agent.firewall_refresh_needed()):
                LOG.debug("Agent loop found changes! %s", device_info)
                try:
                    sync = self.process_network_devices(device_info)
                except Exception:
                    LOG.exception("Error in agent loop. Devices info: %s",
                                  device_info)
                    sync = True

            # sleep till end of polling interval
            elapsed = (time.time() - start)
            if (elapsed < self.polling_interval):
                time.sleep(self.polling_interval - elapsed)
            else:
                LOG.debug("Loop iteration exceeded interval "
                          "(%(polling_interval)s vs. %(elapsed)s)!",
                          {'polling_interval': self.polling_interval,
                           'elapsed': elapsed})

    def set_rpc_timeout(self, timeout):
        for rpc_api in (self.plugin_rpc, self.sg_plugin_rpc,
                        self.state_rpc):
            rpc_api.client.timeout = timeout



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron_lib import constants as n_const
from neutron_lib import context
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_policy import policy as oslo_policy
from oslo_serialization import jsonutils
from oslo_utils import uuidutils
import pecan
from pecan import request

from neutron.api import extensions
from neutron.conf import quota as qconf
from neutron import manager
from neutron.pecan_wsgi.controllers import root as controllers
from neutron.pecan_wsgi.controllers import utils as controller_utils
from neutron import policy
from neutron.tests.common import helpers
from neutron.tests.functional.pecan_wsgi import test_functional
from neutron.tests.functional.pecan_wsgi import utils as pecan_utils
from neutron.tests.unit import dummy_plugin


_SERVICE_PLUGIN_RESOURCE = 'serviceplugin'
_SERVICE_PLUGIN_COLLECTION = _SERVICE_PLUGIN_RESOURCE + 's'
_SERVICE_PLUGIN_INDEX_BODY = {_SERVICE_PLUGIN_COLLECTION: []}


class FakeServicePluginController(controller_utils.NeutronPecanController):
    resource = _SERVICE_PLUGIN_RESOURCE
    collection = _SERVICE_PLUGIN_COLLECTION

    @pecan.expose(generic=True,
                  content_type='application/json',
                  template='json')
    def index(self):
        return _SERVICE_PLUGIN_INDEX_BODY


class TestRootController(test_functional.PecanFunctionalTest):
    """Test version listing on root URI."""

    base_url = '/'

    def setUp(self):
        super(TestRootController, self).setUp()
        self.setup_service_plugin()
        self.plugin = directory.get_plugin()
        self.ctx = context.get_admin_context()

    def setup_service_plugin(self):
        manager.NeutronManager.set_controller_for_resource(
            _SERVICE_PLUGIN_COLLECTION,
            FakeServicePluginController(_SERVICE_PLUGIN_COLLECTION,
                                        _SERVICE_PLUGIN_RESOURCE))

    def _test_method_returns_code(self, method, code=200):
        api_method = getattr(self.app, method)
        response = api_method(self.base_url, expect_errors=True)
        self.assertEqual(response.status_int, code)

    def test_get(self):
        response = self.app.get(self.base_url)
        self.assertEqual(response.status_int, 200)
        json_body = jsonutils.loads(response.body)
        versions = json_body.get('versions')
        self.assertEqual(1, len(versions))
        for (attr, value) in controllers.V2Controller.version_info.items():
            self.assertIn(attr, versions[0])
            self.assertEqual(value, versions[0][attr])

    def test_methods(self):
        self._test_method_returns_code('post', 405)
        self._test_method_returns_code('patch', 405)
        self._test_method_returns_code('delete', 405)
        self._test_method_returns_code('head', 405)
        self._test_method_returns_code('put', 405)


class TestV2Controller(TestRootController):

    base_url = '/v2.0/'

    def test_get(self):
        """Verify current version info are returned."""
        response = self.app.get(self.base_url)
        self.assertEqual(response.status_int, 200)
        json_body = jsonutils.loads(response.body)
        self.assertIn('resources', json_body)
        self.assertIsInstance(json_body['resources'], list)
        for r in json_body['resources']:
            self.assertIn("links", r)
            self.assertIn("name", r)
            self.assertIn("collection", r)
            self.assertIn(self.base_url, r['links'][0]['href'])

    def test_get_no_trailing_slash(self):
        response = self.app.get(self.base_url[:-1], expect_errors=True)
        self.assertEqual(response.status_int, 404)

    def test_routing_successs(self):
        """Test dispatch to controller for existing resource."""
        response = self.app.get('%sports.json' % self.base_url)
        self.assertEqual(response.status_int, 200)

    def test_routing_failure(self):
        """Test dispatch to controller for non-existing resource."""
        response = self.app.get('%sidonotexist.json' % self.base_url,
                                expect_errors=True)
        self.assertEqual(response.status_int, 404)

    def test_methods(self):
        self._test_method_returns_code('post', 405)
        self._test_method_returns_code('put', 405)
        self._test_method_returns_code('patch', 405)
        self._test_method_returns_code('delete', 405)
        self._test_method_returns_code('head', 405)
        self._test_method_returns_code('delete', 405)


class TestExtensionsController(TestRootController):
    """Test extension listing and detail reporting."""

    base_url = '/v2.0/extensions'

    def _get_supported_extensions(self):
        ext_mgr = extensions.PluginAwareExtensionManager.get_instance()
        return ext_mgr.get_supported_extension_aliases()

    def test_index(self):
        response = self.app.get(self.base_url)
        self.assertEqual(response.status_int, 200)
        json_body = jsonutils.loads(response.body)
        returned_aliases = [ext['alias'] for ext in json_body['extensions']]
        supported_extensions = self._get_supported_extensions()
        self.assertEqual(supported_extensions, set(returned_aliases))

    def test_get(self):
        # Fetch any extension supported by plugins
        test_alias = self._get_supported_extensions().pop()
        response = self.app.get('%s/%s' % (self.base_url, test_alias))
        self.assertEqual(response.status_int, 200)
        json_body = jsonutils.loads(response.body)
        self.assertEqual(test_alias, json_body['extension']['alias'])

    def test_methods(self):
        self._test_method_returns_code('post', 404)
        self._test_method_returns_code('put', 404)
        self._test_method_returns_code('patch', 404)
        self._test_method_returns_code('delete', 404)
        self._test_method_returns_code('head', 404)
        self._test_method_returns_code('delete', 404)


class TestQuotasController(test_functional.PecanFunctionalTest):
    """Test quota management API controller."""

    base_url = '/v2.0/quotas'
    default_expected_limits = {
        'network': qconf.DEFAULT_QUOTA_NETWORK,
        'port': qconf.DEFAULT_QUOTA_PORT,
        'subnet': qconf.DEFAULT_QUOTA_SUBNET}

    def _verify_limits(self, response, limits):
        for resource, limit in limits.items():
            self.assertEqual(limit, response['quota'][resource])

    def _verify_default_limits(self, response):
        self._verify_limits(response, self.default_expected_limits)

    def _verify_after_update(self, response, updated_limits):
        expected_limits = self.default_expected_limits.copy()
        expected_limits.update(updated_limits)
        self._verify_limits(response, expected_limits)

    def test_index_admin(self):
        # NOTE(salv-orlando): The quota controller has an hardcoded check for
        # admin-ness for this operation, which is supposed to return quotas for
        # all tenants. Such check is "vestigial" from the home-grown WSGI and
        # shall be removed
        response = self.app.get('%s.json' % self.base_url,
                                headers={'X-Project-Id': 'admin',
                                         'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)

    def test_index(self):
        response = self.app.get('%s.json' % self.base_url, expect_errors=True)
        self.assertEqual(403, response.status_int)

    def test_get_admin(self):
        response = self.app.get('%s/foo.json' % self.base_url,
                                headers={'X-Project-Id': 'admin',
                                         'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)
        # As quota limits have not been updated, expect default values
        json_body = jsonutils.loads(response.body)
        self._verify_default_limits(json_body)

    def test_get(self):
        # It is not ok to access another tenant's limits
        url = '%s/foo.json' % self.base_url
        response = self.app.get(url, expect_errors=True)
        self.assertEqual(403, response.status_int)
        # It is however ok to retrieve your own limits
        response = self.app.get(url, headers={'X-Project-Id': 'foo'})
        self.assertEqual(200, response.status_int)
        json_body = jsonutils.loads(response.body)
        self._verify_default_limits(json_body)

    def test_put_get_delete(self):
        # PUT and DELETE actions are in the same test as a meaningful DELETE
        # test would require a put anyway
        url = '%s/foo.json' % self.base_url
        response = self.app.put_json(url,
                                     params={'quota': {'network': 99}},
                                     headers={'X-Project-Id': 'admin',
                                              'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)
        json_body = jsonutils.loads(response.body)
        self._verify_after_update(json_body, {'network': 99})

        response = self.app.get(url, headers={'X-Project-Id': 'foo'})
        self.assertEqual(200, response.status_int)
        json_body = jsonutils.loads(response.body)
        self._verify_after_update(json_body, {'network': 99})

        response = self.app.delete(url, headers={'X-Project-Id': 'admin',
                                                 'X-Roles': 'admin'})
        self.assertEqual(204, response.status_int)
        self.assertFalse(response.body)
        # As DELETE does not return a body we need another GET
        response = self.app.get(url, headers={'X-Project-Id': 'foo'})
        self.assertEqual(200, response.status_int)
        json_body = jsonutils.loads(response.body)
        self._verify_default_limits(json_body)

    def test_update_list_delete(self):
        # PUT and DELETE actions are in the same test as a meaningful DELETE
        # test would require a put anyway
        url = '%s/foo.json' % self.base_url
        response = self.app.put_json(url,
                                     params={'quota': {'network': 99}},
                                     headers={'X-Project-Id': 'admin',
                                              'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)
        json_body = jsonutils.loads(response.body)
        self._verify_after_update(json_body, {'network': 99})

        response = self.app.get(self.base_url,
                                headers={'X-Project-Id': 'admin',
                                         'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)
        json_body = jsonutils.loads(response.body)
        found = False
        for qs in json_body['quotas']:
            if qs['tenant_id'] == 'foo':
                found = True
        self.assertTrue(found)

        response = self.app.delete(url, headers={'X-Project-Id': 'admin',
                                                 'X-Roles': 'admin'})
        self.assertEqual(204, response.status_int)
        self.assertFalse(response.body)
        response = self.app.get(self.base_url,
                                headers={'X-Project-Id': 'admin',
                                         'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)
        json_body = jsonutils.loads(response.body)
        for qs in json_body['quotas']:
            self.assertNotEqual('foo', qs['tenant_id'])

    def test_quotas_get_defaults(self):
        response = self.app.get('%s/foo/default.json' % self.base_url,
                                headers={'X-Project-Id': 'admin',
                                         'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)
        # As quota limits have not been updated, expect default values
        json_body = jsonutils.loads(response.body)
        self._verify_default_limits(json_body)

    def test_get_tenant_info(self):
        response = self.app.get('%s/tenant.json' % self.base_url,
                                headers={'X-Project-Id': 'admin',
                                         'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)
        json_body = jsonutils.loads(response.body)
        self.assertEqual('admin', json_body['tenant']['tenant_id'])


class TestResourceController(TestRootController):
    """Test generic controller"""
    # TODO(salv-orlando): This test case must not explicitly test the 'port'
    # resource. Also it should validate correct plugin/resource association
    base_url = '/v2.0'

    def setUp(self):
        super(TestResourceController, self).setUp()
        policy.init()
        self.addCleanup(policy.reset)
        self._gen_port()

    def _gen_port(self):
        network_id = self.plugin.create_network(context.get_admin_context(), {
            'network':
            {'name': 'pecannet', 'tenant_id': 'tenid', 'shared': False,
             'admin_state_up': True, 'status': 'ACTIVE'}})['id']
        self.port = self.plugin.create_port(context.get_admin_context(), {
            'port':
            {'tenant_id': 'tenid', 'network_id': network_id,
             'fixed_ips': n_const.ATTR_NOT_SPECIFIED,
             'mac_address': '00:11:22:33:44:55',
             'admin_state_up': True, 'device_id': 'FF',
             'device_owner': 'pecan', 'name': 'pecan'}})

    def test_get(self):
        response = self.app.get('/v2.0/ports.json')
        self.assertEqual(response.status_int, 200)

    def _check_item(self, expected, item):
        for attribute in expected:
            self.assertIn(attribute, item)
        self.assertEqual(len(expected), len(item))

    def _test_get_collection_with_fields_selector(self, fields=None):
        fields = fields or []
        query_params = ['fields=%s' % field for field in fields]
        url = '/v2.0/ports.json'
        if query_params:
            url = '%s?%s' % (url, '&'.join(query_params))
        list_resp = self.app.get(url, headers={'X-Project-Id': 'tenid'})
        self.assertEqual(200, list_resp.status_int)
        for item in jsonutils.loads(list_resp.body).get('ports', []):
            for field in fields:
                self.assertIn(field, item)
            if fields:
                self.assertEqual(len(fields), len(item))
            else:
                for field in ('id', 'name', 'device_owner'):
                    self.assertIn(field, item)

    def test_get_collection_with_multiple_fields_selector(self):
        self._test_get_collection_with_fields_selector(fields=['id', 'name'])

    def test_get_collection_with_single_fields_selector(self):
        self._test_get_collection_with_fields_selector(fields=['name'])

    def test_get_collection_without_fields_selector(self):
        self._test_get_collection_with_fields_selector(fields=[])

    def test_project_id_in_mandatory_fields(self):
        # ports only specifies that tenant_id is mandatory, but project_id
        # should still be passed to the plugin.
        mock_get = mock.patch.object(self.plugin, 'get_ports',
                                     return_value=[]).start()
        self.app.get(
            '/v2.0/ports.json?fields=id',
            headers={'X-Project-Id': 'tenid'}
        )
        self.assertIn('project_id', mock_get.mock_calls[-1][2]['fields'])

    def test_get_item_with_fields_selector(self):
        item_resp = self.app.get(
            '/v2.0/ports/%s.json?fields=id&fields=name' % self.port['id'],
            headers={'X-Project-Id': 'tenid'})
        self.assertEqual(200, item_resp.status_int)
        self._check_item(['id', 'name'],
                         jsonutils.loads(item_resp.body)['port'])
        # Explicitly require an attribute which is also 'required_by_policy'.
        # The attribute should not be stripped while generating the response
        item_resp = self.app.get(
            '/v2.0/ports/%s.json?fields=id&fields=tenant_id' % self.port['id'],
            headers={'X-Project-Id': 'tenid'})
        self.assertEqual(200, item_resp.status_int)
        self._check_item(['id', 'tenant_id'],
                         jsonutils.loads(item_resp.body)['port'])

    def test_duped_and_empty_fields_stripped(self):
        mock_get = mock.patch.object(self.plugin, 'get_ports',
                                     return_value=[]).start()
        self.app.get(
            '/v2.0/ports.json?fields=id&fields=name&fields=&fields=name',
            headers={'X-Project-Id': 'tenid'}
        )
        received = mock_get.mock_calls[-1][2]['fields']
        self.assertNotIn('', received)
        self.assertEqual(len(received), len(set(received)))

    def test_post(self):
        response = self.app.post_json(
            '/v2.0/ports.json',
            params={'port': {'network_id': self.port['network_id'],
                             'admin_state_up': True,
                             'tenant_id': 'tenid'}},
            headers={'X-Project-Id': 'tenid'})
        self.assertEqual(response.status_int, 201)

    def test_post_with_retry(self):
        self._create_failed = False
        orig = self.plugin.create_port

        def new_create(*args, **kwargs):
            if not self._create_failed:
                self._create_failed = True
                raise db_exc.RetryRequest(ValueError())
            return orig(*args, **kwargs)

        with mock.patch.object(self.plugin, 'create_port',
                               new=new_create):
            response = self.app.post_json(
                '/v2.0/ports.json',
                params={'port': {'network_id': self.port['network_id'],
                                 'admin_state_up': True,
                                 'tenant_id': 'tenid'}},
                headers={'X-Project-Id': 'tenid'})
            self.assertEqual(201, response.status_int)

    def test_put(self):
        response = self.app.put_json('/v2.0/ports/%s.json' % self.port['id'],
                                     params={'port': {'name': 'test'}},
                                     headers={'X-Project-Id': 'tenid'})
        self.assertEqual(response.status_int, 200)
        json_body = jsonutils.loads(response.body)
        self.assertEqual(1, len(json_body))
        self.assertIn('port', json_body)
        self.assertEqual('test', json_body['port']['name'])
        self.assertEqual('tenid', json_body['port']['tenant_id'])

    def test_delete(self):
        response = self.app.delete('/v2.0/ports/%s.json' % self.port['id'],
                                   headers={'X-Project-Id': 'tenid'})
        self.assertEqual(response.status_int, 204)
        self.assertFalse(response.body)

    def test_delete_disallows_body(self):
        response = self.app.delete_json(
            '/v2.0/ports/%s.json' % self.port['id'],
            params={'port': {'name': 'test'}},
            headers={'X-Project-Id': 'tenid'},
            expect_errors=True)
        self.assertEqual(response.status_int, 400)

    def test_plugin_initialized(self):
        self.assertIsNotNone(manager.NeutronManager._instance)

    def test_methods(self):
        self._test_method_returns_code('post', 405)
        self._test_method_returns_code('put', 405)
        self._test_method_returns_code('patch', 405)
        self._test_method_returns_code('delete', 405)
        self._test_method_returns_code('head', 405)
        self._test_method_returns_code('delete', 405)

    def test_post_with_empty_body(self):
        response = self.app.post_json(
            '/v2.0/ports.json',
            headers={'X-Project-Id': 'tenid'},
            params={},
            expect_errors=True)
        self.assertEqual(response.status_int, 400)

    def test_post_with_unsupported_json_type(self):
        response = self.app.post_json(
            '/v2.0/ports.json',
            headers={'X-Project-Id': 'tenid'},
            params=[1, 2, 3],
            expect_errors=True)
        self.assertEqual(response.status_int, 400)

    def test_bulk_create(self):
        response = self.app.post_json(
            '/v2.0/ports.json',
            params={'ports': [{'network_id': self.port['network_id'],
                             'admin_state_up': True,
                             'tenant_id': 'tenid'},
                             {'network_id': self.port['network_id'],
                              'admin_state_up': True,
                              'tenant_id': 'tenid'}]
                    },
            headers={'X-Project-Id': 'tenid'})
        self.assertEqual(response.status_int, 201)
        json_body = jsonutils.loads(response.body)
        self.assertIn('ports', json_body)
        self.assertEqual(2, len(json_body['ports']))

    def test_emulated_bulk_create(self):
        self.plugin._FORCE_EMULATED_BULK = True
        response = self.app.post_json(
            '/v2.0/ports.json',
            params={'ports': [{'network_id': self.port['network_id'],
                             'admin_state_up': True,
                             'tenant_id': 'tenid'},
                             {'network_id': self.port['network_id'],
                              'admin_state_up': True,
                              'tenant_id': 'tenid'}]
                    },
            headers={'X-Project-Id': 'tenid'})
        self.assertEqual(response.status_int, 201)
        json_body = jsonutils.loads(response.body)
        self.assertIn('ports', json_body)
        self.assertEqual(2, len(json_body['ports']))

    def test_emulated_bulk_create_rollback(self):
        self.plugin._FORCE_EMULATED_BULK = True
        response = self.app.post_json(
            '/v2.0/ports.json',
            params={'ports': [{'network_id': self.port['network_id'],
                             'admin_state_up': True,
                             'tenant_id': 'tenid'},
                             {'network_id': self.port['network_id'],
                              'admin_state_up': True,
                              'tenant_id': 'tenid'},
                             {'network_id': 'bad_net_id',
                              'admin_state_up': True,
                              'tenant_id': 'tenid'}]
                    },
            headers={'X-Project-Id': 'tenid'},
            expect_errors=True)
        self.assertEqual(response.status_int, 400)
        response = self.app.get(
            '/v2.0/ports.json',
            headers={'X-Project-Id': 'tenid'})
        # all ports should be rolled back from above so we are just left
        # with the one created in setup
        self.assertEqual(1, len(jsonutils.loads(response.body)['ports']))

    def test_bulk_create_one_item(self):
        response = self.app.post_json(
            '/v2.0/ports.json',
            params={'ports': [{'network_id': self.port['network_id'],
                               'admin_state_up': True,
                               'tenant_id': 'tenid'}]
                    },
            headers={'X-Project-Id': 'tenid'})
        self.assertEqual(response.status_int, 201)
        json_body = jsonutils.loads(response.body)
        self.assertIn('ports', json_body)
        self.assertEqual(1, len(json_body['ports']))


class TestPaginationAndSorting(test_functional.PecanFunctionalTest):

    RESOURCE_COUNT = 6

    def setUp(self):
        super(TestPaginationAndSorting, self).setUp()
        policy.init()
        self.addCleanup(policy.reset)
        self.plugin = directory.get_plugin()
        self.ctx = context.get_admin_context()
        self._create_networks(self.RESOURCE_COUNT)
        self.networks = self._get_collection()['networks']

    def _create_networks(self, count=1):
        network_ids = []
        for index in range(count):
            network = {'name': 'pecannet-%d' % index, 'tenant_id': 'tenid',
                       'shared': False, 'admin_state_up': True,
                       'status': 'ACTIVE'}
            network_id = self.plugin.create_network(
                self.ctx, {'network': network})['id']
            network_ids.append(network_id)
        return network_ids

    def _get_collection(self, collection=None, limit=None, marker=None,
                        fields=None, page_reverse=False, sort_key=None,
                        sort_dir=None):
        collection = collection or 'networks'
        fields = fields or []
        query_params = []
        if limit:
            query_params.append('limit=%d' % limit)
        if marker:
            query_params.append('marker=%s' % marker)
        if page_reverse:
            query_params.append('page_reverse=True')
        if sort_key:
            query_params.append('sort_key=%s' % sort_key)
        if sort_dir:
            query_params.append('sort_dir=%s' % sort_dir)
        query_params.extend(['%s%s' % ('fields=', field) for field in fields])
        url = '/v2.0/%s.json' % collection
        if query_params:
            url = '%s?%s' % (url, '&'.join(query_params))
        list_resp = self.app.get(url, headers={'X-Project-Id': 'tenid'})
        self.assertEqual(200, list_resp.status_int)
        return list_resp.json

    def _test_get_collection_with_pagination(self, expected_list,
                                             collection=None,
                                             limit=None, marker=None,
                                             fields=None, page_reverse=False,
                                             sort_key=None, sort_dir=None):
        expected_list = expected_list or []
        collection = collection or 'networks'
        list_resp = self._get_collection(collection=collection, limit=limit,
                                         marker=marker, fields=fields,
                                         page_reverse=page_reverse,
                                         sort_key=sort_key, sort_dir=sort_dir)
        if limit and marker:
            links_key = '%s_links' % collection
            self.assertIn(links_key, list_resp)
        if not fields or 'id' in fields:
            list_resp_ids = [item['id'] for item in list_resp[collection]]
            self.assertEqual(expected_list, list_resp_ids)
        if fields:
            for item in list_resp[collection]:
                for field in fields:
                    self.assertIn(field, item)

    def test_get_collection_with_pagination_limit(self):
        self._test_get_collection_with_pagination([self.networks[0]['id']],
                                                  limit=1)

    def test_get_collection_with_pagination_fields_no_pk(self):
        self._test_get_collection_with_pagination([self.networks[0]['id']],
                                                  limit=1, fields=['name'])

    def test_get_collection_with_pagination_limit_over_count(self):
        expected_ids = [network['id'] for network in self.networks]
        self._test_get_collection_with_pagination(
            expected_ids, limit=self.RESOURCE_COUNT + 1)

    def test_get_collection_with_pagination_marker(self):
        marker = self.networks[2]['id']
        expected_ids = [network['id'] for network in self.networks[3:]]
        self._test_get_collection_with_pagination(expected_ids, limit=3,
                                                  marker=marker)

    def test_get_collection_with_pagination_marker_without_limit(self):
        marker = self.networks[2]['id']
        expected_ids = [network['id'] for network in self.networks]
        self._test_get_collection_with_pagination(expected_ids, marker=marker)

    def test_get_collection_with_pagination_and_fields(self):
        expected_ids = [network['id'] for network in self.networks[:2]]
        self._test_get_collection_with_pagination(
            expected_ids, limit=2, fields=['id', 'name'])

    def test_get_collection_with_pagination_page_reverse(self):
        marker = self.networks[2]['id']
        expected_ids = [network['id'] for network in self.networks[:2]]
        self._test_get_collection_with_pagination(expected_ids, limit=3,
                                                  marker=marker,
                                                  page_reverse=True)

    def test_get_collection_with_sorting_desc(self):
        nets = sorted(self.networks, key=lambda net: net['name'], reverse=True)
        expected_ids = [network['id'] for network in nets]
        self._test_get_collection_with_pagination(expected_ids,
                                                  sort_key='name',
                                                  sort_dir='desc')

    def test_get_collection_with_sorting_asc(self):
        nets = sorted(self.networks, key=lambda net: net['name'])
        expected_ids = [network['id'] for network in nets]
        self._test_get_collection_with_pagination(expected_ids,
                                                  sort_key='name',
                                                  sort_dir='asc')


class TestRequestProcessing(TestRootController):

    def setUp(self):
        super(TestRequestProcessing, self).setUp()
        mock.patch('neutron.pecan_wsgi.hooks.notifier.registry').start()
        # request.context is thread-local storage so it has to be accessed by
        # the controller. We can capture it into a list here to assert on after
        # the request finishes.

        def capture_request_details(*args, **kwargs):
            self.captured_context = request.context
            self.request_params = kwargs

        mock.patch('neutron.pecan_wsgi.controllers.resource.'
                   'CollectionsController.get',
                   side_effect=capture_request_details).start()
        mock.patch('neutron.pecan_wsgi.controllers.resource.'
                   'CollectionsController.create',
                   side_effect=capture_request_details).start()
        mock.patch('neutron.pecan_wsgi.controllers.resource.'
                   'ItemController.get',
                   side_effect=capture_request_details).start()
    # TODO(kevinbenton): add context tests for X-Roles etc

    def test_context_set_in_request(self):
        self.app.get('/v2.0/ports.json',
                     headers={'X-Project-Id': 'tenant_id'})
        self.assertEqual('tenant_id',
                         self.captured_context['neutron_context'].tenant_id)

    def test_core_resource_identified(self):
        self.app.get('/v2.0/ports.json')
        self.assertEqual('port', self.captured_context['resource'])
        self.assertEqual('ports', self.captured_context['collection'])

    def test_lookup_identifies_resource_id(self):
        # We now this will return a 404 but that's not the point as it is
        # mocked
        self.app.get('/v2.0/ports/reina.json')
        self.assertEqual('port', self.captured_context['resource'])
        self.assertEqual('ports', self.captured_context['collection'])
        self.assertEqual('reina', self.captured_context['resource_id'])

    def test_resource_processing_post(self):
        self.app.post_json(
            '/v2.0/networks.json',
            params={'network': {'name': 'the_net',
                                'admin_state_up': True}},
            headers={'X-Project-Id': 'tenid'})
        self.assertEqual('network', self.captured_context['resource'])
        self.assertEqual('networks', self.captured_context['collection'])
        resources = self.captured_context['resources']
        is_bulk = self.captured_context['is_bulk']
        self.assertEqual(1, len(resources))
        self.assertEqual('the_net', resources[0]['name'])
        self.assertTrue(resources[0]['admin_state_up'])
        self.assertFalse(is_bulk)

    def test_resource_processing_post_bulk(self):
        self.app.post_json(
            '/v2.0/networks.json',
            params={'networks': [{'name': 'the_net_1',
                                  'admin_state_up': True},
                                 {'name': 'the_net_2',
                                  'admin_state_up': False}]},
            headers={'X-Project-Id': 'tenid'})
        resources = self.captured_context['resources']
        is_bulk = self.captured_context['is_bulk']
        self.assertEqual(2, len(resources))
        self.assertTrue(resources[0]['admin_state_up'])
        self.assertEqual('the_net_1', resources[0]['name'])
        self.assertFalse(resources[1]['admin_state_up'])
        self.assertEqual('the_net_2', resources[1]['name'])
        self.assertTrue(is_bulk)

    def test_resource_processing_post_bulk_one_item(self):
        self.app.post_json(
            '/v2.0/networks.json',
            params={'networks': [{'name': 'the_net_1',
                                  'admin_state_up': True}]},
            headers={'X-Project-Id': 'tenid'})
        resources = self.captured_context['resources']
        is_bulk = self.captured_context['is_bulk']
        self.assertEqual(1, len(resources))
        self.assertTrue(is_bulk)

    def test_resource_processing_post_unknown_attribute_returns_400(self):
        response = self.app.post_json(
            '/v2.0/networks.json',
            params={'network': {'name': 'the_net',
                                'alien': 'E.T.',
                                'admin_state_up': True}},
            headers={'X-Project-Id': 'tenid'},
            expect_errors=True)
        self.assertEqual(400, response.status_int)

    def test_resource_processing_post_validation_error_returns_400(self):
        response = self.app.post_json(
            '/v2.0/networks.json',
            params={'network': {'name': 'the_net',
                                'admin_state_up': 'invalid_value'}},
            headers={'X-Project-Id': 'tenid'},
            expect_errors=True)
        self.assertEqual(400, response.status_int)

    def test_service_plugin_identified(self):
        # TODO(kevinbenton): fix the unit test setup to include an l3 plugin
        self.skipTest("A dummy l3 plugin needs to be setup")
        self.app.get('/v2.0/routers.json')
        self.assertEqual('router', self.req_stash['resource_type'])
        # make sure the core plugin was identified as the handler for ports
        self.assertEqual(
            directory.get_plugin(plugin_constants.L3),
            self.req_stash['plugin'])

    def test_service_plugin_uri(self):
        nm = manager.NeutronManager.get_instance()
        nm.path_prefix_resource_mappings[dummy_plugin.RESOURCE_NAME] = [
            _SERVICE_PLUGIN_COLLECTION]
        response = self.do_request('/v2.0/dummy/serviceplugins.json')
        self.assertEqual(200, response.status_int)
        self.assertEqual(_SERVICE_PLUGIN_INDEX_BODY, response.json_body)


class TestRouterController(TestResourceController):
    """Specialized tests for the router resource controller

    This test class adds tests specific for the router controller in
    order to verify the 'member_action' functionality, which this
    controller uses for adding and removing router interfaces.
    """

    def setUp(self):
        cfg.CONF.set_override(
            'service_plugins',
            ['neutron.services.l3_router.l3_router_plugin.L3RouterPlugin',
             'neutron.services.flavors.flavors_plugin.FlavorsPlugin'])
        super(TestRouterController, self).setUp()
        policy.init()
        self.addCleanup(policy.reset)
        plugin = directory.get_plugin()
        ctx = context.get_admin_context()
        l3_plugin = directory.get_plugin(plugin_constants.L3)
        network_id = pecan_utils.create_network(ctx, plugin)['id']
        self.subnet = pecan_utils.create_subnet(ctx, plugin, network_id)
        self.router = pecan_utils.create_router(ctx, l3_plugin)

    def test_member_actions_processing(self):
        response = self.app.put_json(
            '/v2.0/routers/%s/add_router_interface.json' % self.router['id'],
            params={'subnet_id': self.subnet['id']},
            headers={'X-Project-Id': 'tenid'})
        self.assertEqual(200, response.status_int)

    def test_non_existing_member_action_returns_404(self):
        response = self.app.put_json(
            '/v2.0/routers/%s/do_meh.json' % self.router['id'],
            params={'subnet_id': 'doesitevenmatter'},
            headers={'X-Project-Id': 'tenid'},
            expect_errors=True)
        self.assertEqual(404, response.status_int)

    def test_unsupported_method_member_action(self):
        response = self.app.post_json(
            '/v2.0/routers/%s/add_router_interface.json' % self.router['id'],
            params={'subnet_id': self.subnet['id']},
            headers={'X-Project-Id': 'tenid'},
            expect_errors=True)
        self.assertEqual(405, response.status_int)

        response = self.app.get(
            '/v2.0/routers/%s/add_router_interface.json' % self.router['id'],
            headers={'X-Project-Id': 'tenid'},
            expect_errors=True)
        self.assertEqual(405, response.status_int)


class TestDHCPAgentShimControllers(test_functional.PecanFunctionalTest):

    def setUp(self):
        super(TestDHCPAgentShimControllers, self).setUp()
        policy.init()
        policy._ENFORCER.set_rules(
            oslo_policy.Rules.from_dict(
                {'get_dhcp-agents': 'role:admin',
                 'get_dhcp-networks': 'role:admin',
                 'create_dhcp-networks': 'role:admin',
                 'delete_dhcp-networks': 'role:admin'}),
            overwrite=False)
        plugin = directory.get_plugin()
        ctx = context.get_admin_context()
        self.network = pecan_utils.create_network(ctx, plugin)
        self.agent = helpers.register_dhcp_agent()
        # NOTE(blogan): Not sending notifications because this test is for
        # testing the shim controllers
        plugin.agent_notifiers[n_const.AGENT_TYPE_DHCP] = None

    def test_list_dhcp_agents_hosting_network(self):
        response = self.app.get(
            '/v2.0/networks/%s/dhcp-agents.json' % self.network['id'],
            headers={'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)

    def test_list_networks_on_dhcp_agent(self):
        response = self.app.get(
            '/v2.0/agents/%s/dhcp-networks.json' % self.agent.id,
            headers={'X-Project-Id': 'tenid', 'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)

    def test_add_remove_dhcp_agent(self):
        headers = {'X-Project-Id': 'tenid', 'X-Roles': 'admin'}
        self.app.post_json(
            '/v2.0/agents/%s/dhcp-networks.json' % self.agent.id,
            headers=headers, params={'network_id': self.network['id']})
        response = self.app.get(
            '/v2.0/networks/%s/dhcp-agents.json' % self.network['id'],
            headers=headers)
        self.assertIn(self.agent.id,
                      [a['id'] for a in response.json['agents']])
        self.app.delete('/v2.0/agents/%(a)s/dhcp-networks/%(n)s.json' % {
            'a': self.agent.id, 'n': self.network['id']}, headers=headers)
        response = self.app.get(
            '/v2.0/networks/%s/dhcp-agents.json' % self.network['id'],
            headers=headers)
        self.assertNotIn(self.agent.id,
                         [a['id'] for a in response.json['agents']])


class TestL3AgentShimControllers(test_functional.PecanFunctionalTest):

    def setUp(self):
        cfg.CONF.set_override(
            'service_plugins',
            ['neutron.services.l3_router.l3_router_plugin.L3RouterPlugin',
             'neutron.services.flavors.flavors_plugin.FlavorsPlugin'])
        super(TestL3AgentShimControllers, self).setUp()
        policy.init()
        policy._ENFORCER.set_rules(
            oslo_policy.Rules.from_dict(
                {'get_l3-agents': 'role:admin',
                 'get_l3-routers': 'role:admin'}),
            overwrite=False)
        ctx = context.get_admin_context()
        l3_plugin = directory.get_plugin(plugin_constants.L3)
        self.router = pecan_utils.create_router(ctx, l3_plugin)
        self.agent = helpers.register_l3_agent()
        # NOTE(blogan): Not sending notifications because this test is for
        # testing the shim controllers
        l3_plugin.agent_notifiers[n_const.AGENT_TYPE_L3] = None

    def test_list_l3_agents_hosting_router(self):
        response = self.app.get(
            '/v2.0/routers/%s/l3-agents.json' % self.router['id'],
            headers={'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)

    def test_list_routers_on_l3_agent(self):
        response = self.app.get(
            '/v2.0/agents/%s/l3-routers.json' % self.agent.id,
            headers={'X-Roles': 'admin'})
        self.assertEqual(200, response.status_int)

    def test_add_remove_l3_agent(self):
        headers = {'X-Project-Id': 'tenid', 'X-Roles': 'admin'}
        response = self.app.post_json(
            '/v2.0/agents/%s/l3-routers.json' % self.agent.id,
            headers=headers, params={'router_id': self.router['id']})
        self.assertEqual(201, response.status_int)
        response = self.app.get(
            '/v2.0/routers/%s/l3-agents.json' % self.router['id'],
            headers=headers)
        self.assertIn(self.agent.id,
                      [a['id'] for a in response.json['agents']])
        response = self.app.delete(
            '/v2.0/agents/%(a)s/l3-routers/%(n)s.json' % {
                'a': self.agent.id, 'n': self.router['id']}, headers=headers)
        self.assertEqual(204, response.status_int)
        self.assertFalse(response.body)
        response = self.app.get(
            '/v2.0/routers/%s/l3-agents.json' % self.router['id'],
            headers=headers)
        self.assertNotIn(self.agent.id,
                         [a['id'] for a in response.json['agents']])


class TestShimControllers(test_functional.PecanFunctionalTest):

    def setUp(self):
        fake_ext = pecan_utils.FakeExtension()
        fake_plugin = pecan_utils.FakePlugin()
        plugins = {pecan_utils.FakePlugin.PLUGIN_TYPE: fake_plugin}
        new_extensions = {fake_ext.get_alias(): fake_ext}
        super(TestShimControllers, self).setUp(
            service_plugins=plugins, extensions=new_extensions)
        policy.init()
        policy._ENFORCER.set_rules(
            oslo_policy.Rules.from_dict(
                {'get_meh_meh': '',
                 'get_meh_mehs': '',
                 'get_fake_subresources': ''}),
            overwrite=False)
        self.addCleanup(policy.reset)

    def test_hyphenated_resource_controller_not_shimmed(self):
        collection = pecan_utils.FakeExtension.HYPHENATED_COLLECTION.replace(
            '_', '-')
        resource = pecan_utils.FakeExtension.HYPHENATED_RESOURCE
        url = '/v2.0/{}/something.json'.format(collection)
        resp = self.app.get(url)
        self.assertEqual(200, resp.status_int)
        self.assertEqual({resource: {'fake': 'something'}}, resp.json)

    def test_hyphenated_collection_controller_not_shimmed(self):
        body_collection = pecan_utils.FakeExtension.HYPHENATED_COLLECTION
        uri_collection = body_collection.replace('_', '-')
        url = '/v2.0/{}.json'.format(uri_collection)
        resp = self.app.get(url)
        self.assertEqual(200, resp.status_int)
        self.assertEqual({body_collection: [{'fake': 'fake'}]}, resp.json)

    def test_hyphenated_collection_subresource_controller_not_shimmed(self):
        body_collection = pecan_utils.FakeExtension.HYPHENATED_COLLECTION
        uri_collection = body_collection.replace('_', '-')
        # there is only one subresource so far
        sub_resource_collection = (
            pecan_utils.FakeExtension.FAKE_SUB_RESOURCE_COLLECTION)
        temp_id = uuidutils.generate_uuid()
        url = '/v2.0/{0}/{1}/{2}'.format(
            uri_collection,
            temp_id,
            sub_resource_collection.replace('_', '-'))
        resp = self.app.get(url)
        self.assertEqual(200, resp.status_int)
        self.assertEqual({sub_resource_collection: {'foo': temp_id}},
                         resp.json)


class TestMemberActionController(test_functional.PecanFunctionalTest):
    def setUp(self):
        fake_ext = pecan_utils.FakeExtension()
        fake_plugin = pecan_utils.FakePlugin()
        plugins = {pecan_utils.FakePlugin.PLUGIN_TYPE: fake_plugin}
        new_extensions = {fake_ext.get_alias(): fake_ext}
        super(TestMemberActionController, self).setUp(
            service_plugins=plugins, extensions=new_extensions)
        hyphen_collection = pecan_utils.FakeExtension.HYPHENATED_COLLECTION
        self.collection = hyphen_collection.replace('_', '-')

    def test_get_member_action_controller(self):
        url = '/v2.0/{}/something/boo_meh.json'.format(self.collection)
        resp = self.app.get(url)
        self.assertEqual(200, resp.status_int)
        self.assertEqual({'boo_yah': 'something'}, resp.json)

    def test_put_member_action_controller(self):
        url = '/v2.0/{}/something/put_meh.json'.format(self.collection)
        resp = self.app.put_json(url, params={'it_matters_not': 'ok'})
        self.assertEqual(200, resp.status_int)
        self.assertEqual({'poo_yah': 'something'}, resp.json)

    def test_get_member_action_does_not_exist(self):
        url = '/v2.0/{}/something/are_you_still_there.json'.format(
            self.collection)
        resp = self.app.get(url, expect_errors=True)
        self.assertEqual(404, resp.status_int)

    def test_put_member_action_does_not_exist(self):
        url = '/v2.0/{}/something/are_you_still_there.json'.format(
            self.collection)
        resp = self.app.put_json(url, params={'it_matters_not': 'ok'},
                                 expect_errors=True)
        self.assertEqual(404, resp.status_int)

    def test_put_on_get_member_action(self):
        url = '/v2.0/{}/something/boo_meh.json'.format(self.collection)
        resp = self.app.put_json(url, params={'it_matters_not': 'ok'},
                                 expect_errors=True)
        self.assertEqual(405, resp.status_int)

    def test_get_on_put_member_action(self):
        url = '/v2.0/{}/something/put_meh.json'.format(self.collection)
        resp = self.app.get(url, expect_errors=True)
        self.assertEqual(405, resp.status_int)


class TestParentSubresourceController(test_functional.PecanFunctionalTest):
    def setUp(self):
        fake_ext = pecan_utils.FakeExtension()
        fake_plugin = pecan_utils.FakePlugin()
        plugins = {pecan_utils.FakePlugin.PLUGIN_TYPE: fake_plugin}
        new_extensions = {fake_ext.get_alias(): fake_ext}
        super(TestParentSubresourceController, self).setUp(
            service_plugins=plugins, extensions=new_extensions)
        policy.init()
        policy._ENFORCER.set_rules(
            oslo_policy.Rules.from_dict(
                {'get_fake_duplicate': '',
                 'get_meh_meh_fake_duplicate': ''}),
            overwrite=False)
        self.addCleanup(policy.reset)
        hyphen_collection = pecan_utils.FakeExtension.HYPHENATED_COLLECTION
        self.collection = hyphen_collection.replace('_', '-')
        self.fake_collection = (pecan_utils.FakeExtension.
                                FAKE_PARENT_SUBRESOURCE_COLLECTION)

    def test_get_duplicate_parent_resource(self):
        url = '/v2.0/{}'.format(self.fake_collection)
        resp = self.app.get(url)
        self.assertEqual(200, resp.status_int)
        self.assertEqual({'fake_duplicates': [{'fake': 'fakeduplicates'}]},
                         resp.json)

    def test_get_duplicate_parent_resource_item(self):
        url = '/v2.0/{}/something'.format(self.fake_collection)
        resp = self.app.get(url)
        self.assertEqual(200, resp.status_int)
        self.assertEqual({'fake_duplicate': {'fake': 'something'}}, resp.json)

    def test_get_parent_resource_and_duplicate_subresources(self):
        url = '/v2.0/{0}/something/{1}'.format(self.collection,
                                               self.fake_collection)
        resp = self.app.get(url)
        self.assertEqual(200, resp.status_int)
        self.assertEqual({'fake_duplicates': [{'fake': 'something'}]},
                         resp.json)

    def test_get_child_resource_policy_check(self):
        policy.reset()
        policy.init()
        policy._ENFORCER.set_rules(
            oslo_policy.Rules.from_dict(
                {'get_meh_meh_fake_duplicate': ''}
            )
        )
        url = '/v2.0/{0}/something/{1}'.format(self.collection,
                                               self.fake_collection)
        resp = self.app.get(url)
        self.assertEqual(200, resp.status_int)
        self.assertEqual({'fake_duplicates': [{'fake': 'something'}]},
                         resp.json)


class TestExcludeAttributePolicy(test_functional.PecanFunctionalTest):

    def setUp(self):
        super(TestExcludeAttributePolicy, self).setUp()
        policy.init()
        self.addCleanup(policy.reset)
        plugin = directory.get_plugin()
        ctx = context.get_admin_context()
        self.network_id = pecan_utils.create_network(ctx, plugin)['id']
        mock.patch('neutron.pecan_wsgi.controllers.resource.'
                   'CollectionsController.get').start()

    def test_get_networks(self):
        response = self.app.get('/v2.0/networks/%s.json' % self.network_id,
                headers={'X-Project-Id': 'tenid'})
        json_body = jsonutils.loads(response.body)
        self.assertEqual(response.status_int, 200)
        self.assertEqual('tenid', json_body['network']['project_id'])
        self.assertEqual('tenid', json_body['network']['tenant_id'])
